# Copyright (c) 2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check functionality of the HWPSS simulation

"""

import unittest

import astropy.units as u
import numpy as np

try:
    # Import sotodlib toast module first, which sets global toast defaults
    import toast
    import toast.ops
    from toast.observation import default_values as defaults

    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    toast_available = True
except ImportError as e:
    toast_available = False

from ._helpers import (calibration_schedule, close_data_and_comm,
                      simulation_test_data)

from sotodlib.toast.ops import pos_to_chi


class SimMuMUXCrosstalkTest(unittest.TestCase):

    def test_sim_hwpss(self):
        if not toast_available:
            print("toast cannot be imported- skipping unit tests", flush=True)
            return

        comm, procs, rank = toast.get_world()
        data = simulation_test_data(
            comm,
            telescope_name="SAT1",
            wafer_slot="w28",
            bands="SAT_f090",
            sample_rate=10.0 * u.Hz,
            thin_fp=1,
            cal_schedule=False,
        )

        pointing = toast.ops.PointingDetectorSimple(
            name="det_pointing_radec",
            quats="quats_radec",
            boresight=defaults.boresight_radec,
            shared_flag_mask=0,
        )

        weights = toast.ops.StokesWeights(
            name="weights_radec",
            weights="weights_radec",
            mode="IQU",
            detector_pointing=pointing,
        )

        # Simple test just confirms that the operator functions

        so_ops.SimHWPSS(
            name="sim_hwpss", stokes_weights=weights
        ).apply(data)

        # Make sure the magnitude of the effect is not crazy

        signal = data.obs[0].detdata["signal"].data.copy()

        rms = np.std(signal)

        assert rms > .1
        assert rms < 10

        close_data_and_comm(data)


    def test_sim_hwpss_drift(self):
        if not toast_available:
            print("toast cannot be imported- skipping unit tests", flush=True)
            return

        comm, procs, rank = toast.get_world()
        data = simulation_test_data(
            comm,
            telescope_name="SAT1",
            wafer_slot="w28",
            bands="SAT_f090",
            sample_rate=10.0 * u.Hz,
            thin_fp=1,
            cal_schedule=False,
        )

        pointing = toast.ops.PointingDetectorSimple(
            name="det_pointing_radec",
            quats="quats_radec",
            boresight=defaults.boresight_radec,
            shared_flag_mask=0,
        )

        weights = toast.ops.StokesWeights(
            name="weights_radec",
            weights="weights_radec",
            mode="IQU",
            detector_pointing=pointing,
        )

        # Simple test just confirms that the operator functions

        so_ops.SimHWPSS(
            name="sim_hwpss", stokes_weights=weights, det_data="no_drift"
        ).apply(data)

        so_ops.SimHWPSS(
            name="sim_hwpss",
            stokes_weights=weights,
            det_data="w_drift",
            drift_rate=0.1,
        ).apply(data)

        # Make sure the magnitude of the effect is not crazy

        no_drift = data.obs[0].detdata["no_drift"].data.copy()
        w_drift = data.obs[0].detdata["w_drift"].data.copy()

        rms = np.std(no_drift)
        rmsdiff = np.std(w_drift - no_drift) / rms

        assert rmsdiff > 1e-3
        assert rmsdiff < 1

        close_data_and_comm(data)

    def test_sim_hwpss_random_drift(self):
        if not toast_available:
            print("toast cannot be imported- skipping unit tests", flush=True)
            return

        comm, procs, rank = toast.get_world()
        data = simulation_test_data(
            comm,
            telescope_name="SAT1",
            wafer_slot="w28",
            bands="SAT_f090",
            sample_rate=10.0 * u.Hz,
            thin_fp=1,
            cal_schedule=False,
        )

        pointing = toast.ops.PointingDetectorSimple(
            name="det_pointing_radec",
            quats="quats_radec",
            boresight=defaults.boresight_radec,
            shared_flag_mask=0,
        )

        weights = toast.ops.StokesWeights(
            name="weights_radec",
            weights="weights_radec",
            mode="IQU",
            detector_pointing=pointing,
        )

        # Simple test just confirms that the operator functions

        so_ops.SimHWPSS(
            name="sim_hwpss", stokes_weights=weights, det_data="no_drift"
        ).apply(data)

        so_ops.SimHWPSS(
            name="sim_hwpss",
            stokes_weights=weights,
            det_data="w_random_drift",
            hwpss_random_drift=True,
            hwpss_drift_alpha=1.0,
            hwpss_drift_coupling_center=1.0,
            hwpss_drift_coupling_width=1e-2,
        ).apply(data)

        # Make sure the magnitude of the effect is not crazy

        no_drift = data.obs[0].detdata["no_drift"].data.copy()
        w_drift = data.obs[0].detdata["w_random_drift"].data.copy()

        rms = np.std(no_drift)
        rmsdiff = np.std(w_drift - no_drift) / rms

        assert rmsdiff > 1e-3
        assert rmsdiff < 10

        close_data_and_comm(data)


if __name__ == '__main__':
    unittest.main()



"""Utility functions for interacting with level 2 data and g3tsmurf
"""

import os
import numpy as np
import logging
import yaml

from sqlalchemy import desc, asc, or_, and_, not_


from so3g.hk import load_range

import sotodlib.core as core
from sotodlib.io.load_smurf import load_file, SmurfStatus, G3tSmurf
from sotodlib.io.g3tsmurf_db import Observations, Files, TimeCodes, SupRsyncType
from sotodlib.io.g3thk_db import G3tHk

from scipy.interpolate import interp1d

logger = logging.getLogger(__name__)


def get_obs_folder(obs_id, archive):
    """
    Get the folder associated with the observation action. Assumes
    everything is following the standard suprsync formatting.
    """
    session = archive.Session()
    obs = session.query(Observations).filter(Observations.obs_id == obs_id).one()
    session.close()

    return os.path.join(
        archive.meta_path,
        str(obs.action_ctime)[:5],
        obs.stream_id,
        str(obs.action_ctime) + "_" + obs.action_name,
    )


def get_obs_outputs(obs_id, archive):
    """
    Get the output files associated with the observation action. Assumes
    everything is following the standard suprsync formatting. Returns
    absolute paths since everything in g3tsmurf is absolute paths.
    """
    path = os.path.join(get_obs_folder(obs_id, archive), "outputs")

    if not os.path.exists(path):
        logger.error(f"Path {path} does not exist, " + f"how does {obs.obs_id} exist?")
    return [os.path.join(path, f) for f in os.listdir(path)]


def get_obs_plots(obs_id, archive):
    """
    Get the output plots associated with the observation action. Assumes
    everything is following the standard suprsync formatting. Returns
    absolute paths since everything in g3tsmurf is absolute paths.
    """
    path = os.path.join(get_obs_folder(obs_id, archive), "plots")

    if not os.path.exists(path):
        logger.error(f"Path {path} does not exist, " + f"how does {obs.obs_id} exist?")
    return [os.path.join(path, f) for f in os.listdir(path)]


def _get_last_oper(timestamp, stream_id, tag, session):
    """Search for observations based on tag, meant to be used to find
    smurf operations as defined by the data packaging setup.
    """
    obs = (
        session.query(Observations)
        .filter(
            Observations.tag.like(tag),
            Observations.timestamp <= timestamp,
            Observations.stream_id == stream_id,
        )
        .order_by(desc(Observations.timestamp))
        .first()
    )
    return obs


def get_last_bg_map(my_obs_id, SMURF):
    """Find the last bias group map relative to a specific observation ID.

    Note: Uses a tag search that was built into sodetlib ~Oct 2022.
    """
    session = SMURF.Session()
    obs = session.query(Observations).filter(Observations.obs_id == my_obs_id).one()
    oper = _get_last_oper(obs.timestamp, obs.stream_id, "oper,bgmap%", session)
    if oper is None:
        logger.error(f"Unable to find bg_map associated with {my_obs_id}")

    files = get_obs_outputs(oper.obs_id, SMURF)
    file = [f for f in files if "bg_map.npy" in f]
    if len(file) != 1:
        logger.error(
            f"Unable to find analysis file in {get_obs_folder(oper.obs_id, SMURF)}"
        )
    session.close()
    return file[0]


def get_last_bias_step(my_obs_id, SMURF):
    """Find the last bias step analysis relative to a specific observation ID.

    Note: Uses a tag search that was built into sodetlib ~Oct 2022.
    """
    session = SMURF.Session()
    obs = session.query(Observations).filter(Observations.obs_id == my_obs_id).one()
    oper = _get_last_oper(obs.timestamp, obs.stream_id, "oper,bias_steps%", session)
    if oper is None:
        logger.error(f"Unable to find Bias Step associated with {my_obs_id}")

    files = get_obs_outputs(oper.obs_id, SMURF)
    file = [f for f in files if "bias_step_analysis.npy" in f]
    if len(file) != 1:
        logger.error(
            f"Unable to find analysis file in {get_obs_folder(oper.obs_id, SMURF)}"
        )
    session.close()
    return file[0]


def get_last_iv(my_obs_id, SMURF):
    """Find the last IV analysis relative to a specific observation ID.

    Note: Uses a tag search that was built into sodetlib ~Oct 2022.
    """
    session = SMURF.Session()
    obs = session.query(Observations).filter(Observations.obs_id == my_obs_id).one()
    oper = _get_last_oper(obs.timestamp, obs.stream_id, "oper,iv%", session)
    if oper is None:
        logger.error(f"Unable to find IV associated with {my_obs_id}")

    files = get_obs_outputs(oper.obs_id, SMURF)
    file = [f for f in files if "iv_analysis.npy" in f]
    if len(file) != 1:
        logger.error(
            f"Unable to find analysis file in {get_obs_folder(oper.obs_id, SMURF)}"
        )
    session.close()
    return file[0]


def _get_next_oper(timestamp, stream_id, tag, session):
    """Search for observations based on tag, meant to be used to find
    smurf operations as defined by the data packaging setup.
    """
    obs = (
        session.query(Observations)
        .filter(
            Observations.tag.like(tag),
            Observations.timestamp >= timestamp,
            Observations.stream_id == stream_id,
        )
        .order_by(asc(Observations.timestamp))
        .first()
    )
    return obs


def get_next_bg_map(my_obs_id, SMURF):
    """Find the next bias group map relative to a specific observation ID.

    Note: Uses a tag search that was built into sodetlib ~Oct 2022.
    """
    session = SMURF.Session()
    obs = session.query(Observations).filter(Observations.obs_id == my_obs_id).one()
    oper = _get_next_oper(obs.timestamp, obs.stream_id, "oper,bgmap%", session)
    if oper is None:
        logger.error(f"Unable to find Bias Step associated with {my_obs_id}")

    files = get_obs_outputs(oper.obs_id, SMURF)
    file = [f for f in files if "bg_map.npy" in f]
    if len(file) != 1:
        logger.error(
            f"Unable to find analysis file in {get_obs_folder(oper.obs_id, SMURF)}"
        )
    session.close()
    return file[0]


def get_next_bias_step(my_obs_id, SMURF):
    """Find the next bias step analysis relative to a specific observation ID.

    Note: Uses a tag search that was built into sodetlib ~Oct 2022.
    """
    session = SMURF.Session()
    obs = session.query(Observations).filter(Observations.obs_id == my_obs_id).one()
    oper = _get_next_oper(obs.timestamp, obs.stream_id, "oper,bias_steps%", session)
    if oper is None:
        logger.error(f"Unable to find Bias Step associated with {my_obs_id}")

    files = get_obs_outputs(oper.obs_id, SMURF)
    file = [f for f in files if "bias_step_analysis.npy" in f]
    if len(file) != 1:
        logger.error(
            f"Unable to find analysis file in {get_obs_folder(oper.obs_id, SMURF)}"
        )
    session.close()
    return file[0]


def get_next_iv(my_obs_id, SMURF):
    """Find the next IV analysis relative to a specific observation ID.

    Note: Uses a tag search that was built into sodetlib ~Oct 2022.
    """
    session = SMURF.Session()
    obs = session.query(Observations).filter(Observations.obs_id == my_obs_id).one()
    oper = _get_next_oper(obs.timestamp, obs.stream_id, "oper,iv%", session)
    if oper is None:
        logger.error(f"Unable to find IV associated with {my_obs_id}")

    files = get_obs_outputs(oper.obs_id, SMURF)
    file = [f for f in files if "iv_analysis.npy" in f]
    if len(file) != 1:
        logger.error(
            f"Unable to find analysis file in {get_obs_folder(oper.obs_id, SMURF)}"
        )
    session.close()
    return file[0]


def check_timecodes(stream_id, start, stop, SMURF):
    """Check if we have timecode entries for a stream_id between
    start and stop timestamps"""
    session = SMURF.Session()
    ## check the timecode dirs first
    t_codes = range(int(start // 1e5), int(stop // 1e5 + 1))
    q = session.query(TimeCodes).filter(
        TimeCodes.stream_id == stream_id,
        or_(*[TimeCodes.timecode == tc for tc in t_codes]),
    )
    files = q.filter(TimeCodes.suprsync_type == SupRsyncType.FILES.value)
    meta = q.filter(TimeCodes.suprsync_type == SupRsyncType.META.value)
    if files.count() == len(t_codes) and meta.count() == len(t_codes):
        return True

    return False


def get_batch(
    obs_id,
    archive,
    ram_limit=None,
    n_det_chunks=None,
    n_samp_chunks=None,
    n_dets=None,
    n_samps=None,
    det_chunks=None,
    samp_chunks=None,
    test=False,
    load_file_args={},
):
    """A Generator to loop through and load AxisManagers of sections of
    Observations. When run with none of the optional arguments it will default
    to returning the full observation. Some arguments over-write others as
    described in the docstrings below. When splitting the Observations by both
    detectors and samples, the chunks of samples for the same detectors are
    looped through first (samples is the inner for loop).

    Example usage::

        for aman in get_batch(obs_id, archive, ram_limit=2e9):
            run_analysis_pipeline(aman)

        for aman in get_batch(obs_id, archive, n_dets=200):
            run_analysis_pipeline(aman)


    Arguments
    ----------
    obs_id : string
        Level 2 observation IDs
    archive : G3tSmurf Instance
        The G3tSmurf database connected to the obs_id
    ram_limit : None or float
        A (very simplistically calculated) limit on RAM per AxisManager. If
        specified it overrides all other inputs for how the AxisManager is
        split.
    n_det_chunks : None or int
        number of chunks of detectors to split the observation by. Each
        AxisManager will have N_det = N_obs_det / n_det_chunks. If specified,
        it overrides n_dets and det_chunks arguments.
    n_samp_chunks: None or int
        number of chunks of samples to split the observation by. Each
        AxisManage will have N_samps = N_obs_samps / n_samps_chunks. If
        specified, it overrides n_samps and samp_chunks arguments.
    n_dets : None or int
        number of detectors to load per AxisManager. If specified, it overrides
        the det_chunks argument.
    n_samps : None or int
        number of samples to load per AxisManager. If specified it overrides
        the samps_chunks arguments.
    det_chunks: None or list of lists, tuples, or ranges
        if specified, each entry in the list is successively passed to load the
        AxisManagers as  `load_file(... channels = list[i] ... )`
    samp_chunks: None or list of tuples
        if specified, each entry in the list is successively passed to load the
        AxisManagers as  `load_file(... samples = list[i] ... )`
    test: bool
        If true, yields a tuple of (det_chunks, samp_chunks) instead of a
        loaded AxisManager
    load_file_kwargs: dict
        additional arguments to pass to load_smurf

    Yields
    --------
    AxisManagers with loaded sections of data
    """

    session = archive.Session()
    obs = session.query(Observations).filter(Observations.obs_id == obs_id).one()
    db_files = session.query(Files).filter(Files.obs_id == obs_id).order_by(Files.start)
    filenames = sorted([f.name for f in db_files])

    # if this throws an error we have some fallbacks
    ts = obs.tunesets[0]
    obs_dets, obs_samps = len(ts.dets), obs.n_samples
    session.close()

    if n_det_chunks is not None and n_dets is not None:
        logger.warning(
            "Both n_det_chunks and n_dets specified, " + "n_det_chunks overrides"
        )
    if n_samp_chunks is not None and n_samps is not None:
        logger.warning(
            "Both n_samp_chunks and n_samps specified, " + "n_samp_chunks overrides"
        )

    logger.debug(f"{obs_id} has (n_dets, n_samps): ({obs_dets}, {obs_samps})")
    if ram_limit is not None:
        pts_limit = int(ram_limit // 4)
        n_samp_chunks = 1
        n_samps = obs_samps
        n_dets = pts_limit // n_samps
        while n_dets == 0:
            n_samp_chunks += 1
            n_samps = obs_samps // n_samp_chunks
            n_dets = pts_limit // (n_samps)
        n_det_chunks = int(np.ceil(obs_dets / n_dets))

    if n_det_chunks is not None:
        n_dets = int(np.ceil(obs_dets / n_det_chunks))
        det_chunks = [
            range(i * n_dets, min((i + 1) * n_dets, obs_dets))
            for i in range(n_det_chunks)
        ]
    if n_samp_chunks is not None:
        n_samps = int(np.ceil(obs_samps / n_samp_chunks))
        samp_chunks = [
            (i * n_samps, min((i + 1) * n_samps, obs_samps))
            for i in range(n_samp_chunks)
        ]

    if n_dets is not None:
        n_det_chunks = int(np.ceil(obs_dets / n_dets))
        det_chunks = [
            range(i * n_dets, min((i + 1) * n_dets, obs_dets))
            for i in range(n_det_chunks)
        ]
    if n_samps is not None:
        n_samp_chunks = int(np.ceil(obs_samps / n_samps))
        samp_chunks = [
            (i * n_samps, min((i + 1) * n_samps, obs_samps))
            for i in range(n_samp_chunks)
        ]

    if det_chunks is None:
        det_chunks = [range(0, obs_dets)]
    if samp_chunks is None:
        samp_chunks = [(0, obs_samps)]

    # should we let folks overwrite this here?
    if "archive" in load_file_args:
        archive = load_file_args.pop("archive")

    if "status" in load_file_args:
        status = load_file_args.pop("status")
    else:
        status = SmurfStatus.from_file(filenames[0])

    logger.debug(f"Loading data with det_chunks: {det_chunks}.")
    logger.debug(f"Loading data in samp_chunks: {samp_chunks}.")

    try:
        for det_chunk in det_chunks:
            for samp_chunk in samp_chunks:
                if test:
                    yield (det_chunk, samp_chunk)
                else:
                    yield load_file(
                        filenames,
                        channels=det_chunk,
                        samples=samp_chunk,
                        archive=archive,
                        status=status,
                        **load_file_args,
                    )
    except GeneratorExit:
        pass


def remove_detmap_info(aman):
    aman.det_info.move("det_id", None)
    aman.det_info.move("wafer", None)


def replace_none(val, dtype, replace_val=np.nan):
    if dtype == str and (val == "null" or val == "NC"):
        return replace_val
    if dtype == float and np.isnan(val) == True:
        return replace_val
    if dtype == int and np.isnan(val) == True:
        return replace_val
    return val


def add_detmap_info(aman, detmap_filename, columns=None):
    """Add detector mapping info into aman.det_info in the format that will be
    generally available once this is done through Context

    Arguments
    ---------
    aman: AxisManager
    detmap_filename : string
        path to file for detmap information. Does just blindly
        assume you are loading the correct file (sorry).
    columns : list of strings
        Optional list of columns to include in addition to main columns.
        Set columns='all' to include all columns.

    """
    detmap = np.genfromtxt(
        detmap_filename,
        delimiter=",",
        skip_header=1,
        dtype=[
            ("smurf_band", "<f8"),
            ("res_index", "<f8"),
            ("freq_mhz", "<f8"),
            ("is_north", "<U5"),
            ("is_highband", "<U5"),
            ("smurf_channel", "<f8"),
            ("smurf_subband", "<f8"),
            ("bond_pad", "<f8"),
            ("mux_band", "<f8"),
            ("mux_channel", "<f8"),
            ("mux_subband", "<U4"),
            ("mux_layout_position", "<f8"),
            ("design_freq_mhz", "<f8"),
            ("bias_line", "<f8"),
            ("pol", "<U4"),
            ("bandpass", "<U4"),
            ("det_row", "<f8"),
            ("det_col", "<f8"),
            ("pixel_num", "<f8"),
            ("pixel_num_pin_skip", "<f8"),
            ("rhomb", "<U4"),
            ("is_optical", "<U5"),
            ("det_x", "<f8"),
            ("det_y", "<f8"),
            ("angle_raw_deg", "<f8"),
            ("angle_actual_deg", "<f8"),
            ("det_type", "<U4"),
            ("detector_id", "<U50"),
            ("flags", "<U50"),
            ("fit_fr_mhz", "<f8"),
            ("fit_q", "<f8"),
            ("fit_qe_real", "<f8"),
            ("fit_qe_imag", "<f8"),
            ("fit_delay_ns", "<f8"),
            ("fit_phi_rad", "<f8"),
            ("fit_fmin_mhz", "<f8"),
            ("fit_amag", "<f8"),
            ("fit_aslope", "<f8"),
        ],
    )

    if "det_id" in aman:
        print("det_id key already in AxisManager, have you already run this function?")
        return
    aman.det_info.wrap_new("det_id", ("dets",), dtype="<U50")

    strings = [
        "is_north",
        "is_highband",
        "mux_subband",
        "rhomb",
        "is_optical",
        "det_type",
        "detector_id",
        "flags",
    ]
    ints = [
        "smurf_band",
        "res_index",
        "smurf_channel",
        "smurf_subband",
        "bond_pad",
        "mux_band",
        "mux_channel",
        "mux_layout_position",
        "bias_line",
        "pixel_num",
        "pixel_num_pin_skip",
    ]
    floats = [
        "freq_mhz",
        "design_freq_mhz",
        "angle_raw_deg",
        "fit_fr_mhz",
        "fit_q",
        "fit_qe_real",
        "fit_qe_imag",
        "fit_delay_ns",
        "fit_phi_rad",
        "fit_fmin_mhz",
        "fit_amag",
        "fit_aslope",
    ]

    wafer = core.AxisManager(aman.dets)
    wafer.wrap_new("pol", ("dets",), dtype="U4")
    wafer.wrap_new("det_x", ("dets",), dtype=float)
    wafer.wrap_new("det_y", ("dets",), dtype=float)
    wafer.wrap_new("angle", ("dets",), dtype=float)
    wafer.wrap_new("det_row", ("dets",), dtype=int)
    wafer.wrap_new("det_col", ("dets",), dtype=int)
    wafer.wrap_new("type", ("dets",), dtype="U4")
    wafer.wrap_new("bandpass", ("dets",), dtype="U4")
    if columns == "all":
        columns = strings + ints + floats
    if columns is not None:
        for col in columns:
            if col in strings:
                wafer.wrap_new(col, ("dets",), dtype="U50")
            elif col in ints:
                wafer.wrap_new(col, ("dets",), dtype=int)
            else:
                wafer.wrap_new(col, ("dets",), dtype=float)

    for i in range(aman.dets.count):
        msk = np.all(
            [
                aman.det_info.smurf.band[i] == detmap["smurf_band"],
                aman.det_info.smurf.channel[i] == detmap["smurf_channel"],
            ],
            axis=0,
        )
        if not np.sum(msk) == 1:
            ## probably fine
            continue

        aman.det_info.det_id[i] = detmap["detector_id"][msk][0]

        wafer["pol"][i] = replace_none(detmap["pol"][msk][0], str, "NC")
        wafer["det_x"][i] = replace_none(detmap["det_x"][msk][0], float, np.nan)
        wafer["det_y"][i] = replace_none(detmap["det_y"][msk][0], float, np.nan)
        wafer["angle"][i] = replace_none(
            np.radians(detmap["angle_actual_deg"][msk][0]), float, np.nan
        )
        wafer["det_row"][i] = replace_none(detmap["det_row"][msk][0], int, -1)
        wafer["det_col"][i] = replace_none(detmap["det_col"][msk][0], int, -1)
        wafer["type"][i] = replace_none(detmap["det_type"][msk][0], str, "NC")
        bandpass = replace_none(detmap["bandpass"][msk][0], str, "NC")
        if not bandpass == "NC":
            bandpass = f"f{bandpass}"
        wafer["bandpass"][i] = bandpass
        if columns is not None:
            for col in columns:
                if col in strings:
                    wafer[col][i] = replace_none(detmap[col][msk][0], str, "NC")
                elif col in ints:
                    wafer[col][i] = replace_none(detmap[col][msk][0], int, -1)
                else:
                    wafer[col][i] = replace_none(detmap[col][msk][0], float, np.nan)

    aman.det_info.wrap("wafer", wafer)


def load_hwp_data(aman, configs=None, data_dir=None):
    """Load HWP data that will match with aman timestamps and
    interpolate angles into hwp_angle

    Arguments
    ----------
    aman : AxisManager
        AxisManager of loaded SMuRF data
    configs : (optional) config filename or dictionary containing "hwp_prefix"
    data_dir : (optional) the "hwp_prefix" data directory. overrides configs

    Wraps HWP data into aman.hwp_angle
    """
    if "hwp_angle" in aman:
        raise ValueError("hwp_angle already exists in aman")

    if configs is None and data_dir is None:
        raise ValueError("Must pass config or data_dir")
    if data_dir is None:
        if type(configs) == str:
            configs = yaml.safe_load(open(configs, "r"))
        data_dir = configs["hwp_prefix"]

    data = load_range(
        float(aman.timestamps[0] - 5),
        float(aman.timestamps[-1] + 5),
        fields=["hwp.hwp_angle"],
        data_dir=data_dir,
    )
    time = data["hwp.hwp_angle"][0]
    if len(time) == 0:
        raise ValueError("No HWP data found that overlaps aman")
    if time[0] > aman.timestamps[0] or time[-1] < aman.timestamps[-1]:
        logger.warning(
            "HWP data does not cover all of AxisManager, "
            + "extrapolations may be unstable"
        )

    hwp_uw = np.unwrap(data["hwp.hwp_angle"][1])
    hwp = interp1d(
        data["hwp.hwp_angle"][0], hwp_uw, bounds_error=False, fill_value="extrapolate"
    )
    aman.wrap("hwp_angle", np.mod(hwp(aman.timestamps), 2 * np.pi), [(0, "samps")])



import sqlalchemy as db
from sqlalchemy.exc import IntegrityError

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship, selectinload

import os
import yaml
import numpy as np
from tqdm import tqdm
from so3g import hk

import logging
from .datapkg_utils import load_configs


logger = logging.getLogger(__name__)

Base = declarative_base()
Session = sessionmaker()


# all database definitions for G3tHK
class HKFiles(Base):
    """This table is named hkfiles and serves as a db for holding
    brief information about each HK'ing file.

    Attributes
    ----------
    filename : string
        name of HK file
    global_start_time : integer
        timestamp that begins a new HK file
    path : string
        path to the HK file
    aggregator : string
        the aggregator corresponding to the data; ex: satp1, satp2, site

    """
    __tablename__ = "hkfiles"
    __table_args__ = (
        db.UniqueConstraint("filename"),
        db.UniqueConstraint("global_start_time"),
    )

    id = db.Column(db.Integer, primary_key=True)

    filename = db.Column(db.String, nullable=False, unique=True)
    global_start_time = db.Column(db.Integer)
    path = db.Column(db.String)
    aggregator = db.Column(db.String)

    fields = relationship("HKFields", back_populates="hkfile")
    agents = relationship("HKAgents", back_populates="hkfile")


class HKAgents(Base):
    """This table is named hkagents; serves as a db for holding
    critical information about each agent in an HK file.

    Attributes
    ----------
    file_id : integer
        id that points an agent back to its corresponding
        HK file in the hkfiles table
    instance_id : string
        the identifier for the instrument/agent; ex: LSA22YG
        for a Lakeshore 372 with the serial number identified as
        an ocs instance-id
    start : integer
        start timestamp corresponding to the agent for corresponding file
    end : integer
        start timestamp corresponding to the agent for corresponding file

    """
    __tablename__ = "hkagents"
    # unique constraint for instance-id? needs more thought?
    id = db.Column(db.Integer, primary_key=True)
    file_id = db.Column(db.Integer, db.ForeignKey("hkfiles.id"))
    hkfile = relationship("HKFiles", back_populates="agents")

    instance_id = db.Column(db.String)

    start = db.Column(db.Integer)
    stop = db.Column(db.Integer)

    fields = relationship("HKFields", back_populates="hkagent")


class HKFields(Base):
    """This table is named hkfields and serves as a db for all
    fields inside each HK file in the hkfeeds table.

    Attributes
    ----------
    file_id : integer
        id that points a field back to its corresponding
        HK file in the hkfiles table
    agent_id : integer
        id that points a field back to its corresponding HK
        agent in the hkagents table
    field : string
        name of HK field in corresponding HK file
    start : integer
        start time for each HK field in ctime for .g3 file
    end : integer
        end time for each HK field in ctime for .g3 file

    """
    __tablename__ = "hkfields"
    id = db.Column(db.Integer, primary_key=True)
    file_id = db.Column(db.Integer, db.ForeignKey("hkfiles.id"))
    hkfile = relationship("HKFiles", back_populates="fields")

    agent_id = db.Column(db.Integer, db.ForeignKey("hkagents.id"))
    hkagent = relationship("HKAgents", back_populates="fields")

    field = db.Column(db.String)
    start = db.Column(db.Integer)
    stop = db.Column(db.Integer)
    median = db.Column(db.Float)
    mean = db.Column(db.Float)
    min_val = db.Column(db.Float)
    max_val = db.Column(db.Float)
    stand_dev = db.Column(db.Float)
    special_math = db.Column(db.Float)


class G3tHk:
    def __init__(self, hkarchive_path, iids, db_path=None, echo=False):
        """
        Class to manage a housekeeping data archive

        Args
        ____
        hkarchive_path : path
            Path to the data directory
        iids : list
            List of agent instance ids
        db_path : path, optional
            Path to the sqlite file
        echo : bool, optional
            If true, all sql statements will print to stdout
        """
        if db_path is None:
            db_path = os.path.join(hkarchive_path, "_HK.db")

        self.hkarchive_path = hkarchive_path
        self.db_path = db_path
        self.iids = iids
        self.engine = db.create_engine(f"sqlite:///{db_path}", echo=echo)
        Session.configure(bind=self.engine)
        self.Session = sessionmaker(bind=self.engine)
        self.session = Session()
        Base.metadata.create_all(self.engine)


    def load_fields(self, hk_path, iids):
        """
        Load fields from .g3 file and start and end time for each field.

        Args
        ____

        hk_path : string
            file path

        returns: list of field names and corresponding start and
                 end times from HK .g3 files
        """

        # enact HKArchiveScanner
        hkas = hk.HKArchiveScanner()
        hkas.process_file(hk_path)
        
        arc = hkas.finalize()

        # get fields from .g3 file
        fields, timelines = arc.get_fields()
        hkfs = []
        for key in fields.keys():
            if any(iid in key for iid in iids):
                hkfs.append(key)

        starts = []
        stops = []
        medians = []
        means = []
        min_vals = []
        max_vals = []
        stds = []

        for i in range(len(hkfs)):
            data = arc.simple(hkfs[i])
            time = data[0]
            starts.append(time[0])
            stops.append(time[-1])
            try:
                medians.append(np.median(data[1]))
                means.append(np.mean(data[1]))
                min_vals.append(np.min(data[1]))
                max_vals.append(np.max(data[1]))
                stds.append(np.std(data[1]))
            except:
                medians.append(np.nan)
                means.append(np.nan)
                min_vals.append(np.nan)
                max_vals.append(np.nan)
                stds.append(np.nan)

        return hkfs, starts, stops, medians, means, min_vals, max_vals, stds

    def _get_agg(self, path_to_file):  # only use hkarchive_path as the FULL path
        """
        Load aggregator information from .g3 file by splitting the field
        names

        Args
        ____

        path_to_file: the full path to a .g3 file which is used in
                      populate_hkfiels() to get the aggregator name

        returns: aggregator name

        """
        # enact HKArchiveScanner
        hkas = hk.HKArchiveScanner()
        hkas.process_file(path_to_file)

        arc = hkas.finalize()

        # get fields from .g3 file
        fields, timelines = arc.get_fields()
        hkfs = []
        for key in fields.keys():
            hkfs.append(key)

        agg_name = hkfs[0].split(".")[0]
        for agg in hkfs:
            assert agg.split(".")[0] == agg_name, "Aggregator name is not consistent"

        return agg_name

    def _files_to_add(self, min_ctime=None, max_ctime=None, update_last_file=True):
        db_files = self.session.query(HKFiles)
        if db_files.count() == 0:
            db_files = []
        else:
            if update_last_file:
                last = db_files.order_by(db.desc(HKFiles.global_start_time)).first()
                db_files = db_files.filter(
                    HKFiles.global_start_time != last.global_start_time
                )
            db_files = db_files.all()
            db_files = [f.path for f in db_files]

        dirs = []
        dir_list = os.listdir(self.hkarchive_path)
        for i in range(len(dir_list)):
            base = os.path.join(self.hkarchive_path, dir_list[i])
            dirs.append(base)
        dirs = sorted(dirs)

        path_list = []

        for i in range(len(dirs)):
            for root, _, files in sorted(os.walk(dirs[i])):
                for f in sorted(files):
                    path = os.path.join(root, f)
                    if f[-3:] != ".g3":
                        continue
                    if path in db_files:
                        continue
                    ftime = int(f.split(".")[0])
                    if min_ctime is not None and ftime < min_ctime:
                        continue
                    if max_ctime is not None and ftime > max_ctime:
                        continue
                    path_list.append(path)

        logger.debug(f"{len(path_list)} files to add to database.")
        return path_list

    def add_hkfiles(
        self,
        min_ctime=None,
        max_ctime=None,
        show_pb=True,
        stop_at_error=False,
        update_last_file=True,
    ):
        """Gather and add column information for hkfiles tables"""

        file_list = self._files_to_add(
            min_ctime=min_ctime,
            max_ctime=max_ctime,
            update_last_file=update_last_file,
        )

        for path in tqdm(sorted(file_list), disable=(not show_pb)):
            try:
                root, filename = os.path.split(path)
                self.add_file(path, overwrite=True)
                self.add_agents_and_fields(path, overwrite=True)

            except IntegrityError as e:
                # Database Integrity Errors, such as duplicate entries
                self.session.rollback()
                logger.warning(f"Integrity error with error {e}")
            except RuntimeError as e:
                # End of stream errors, for G3Files that were not fully flushed
                self.session.rollback()
                logger.warning(f"Failed on file {filename} due to end of stream error!")
            except Exception as e:
                # This will catch generic errors such as attempting to load
                # out-of-date files that do not have the required frame
                # structure specified in the TOD2MAPS docs.
                self.session.rollback()
                if stop_at_error:
                    raise e
                logger.warning(f"Failed on file {filename}:\n{e}")

    def add_file(self, path, overwrite=False):
        db_file = (
            self.session.query(HKFiles)
            .filter(
                HKFiles.path == path,
            )
            .one_or_none()
        )

        if db_file is not None and not overwrite:
            return
        if db_file is None:
            db_file = HKFiles(path=path)
            self.session.add(db_file)

        root, filename = os.path.split(path)
        db_file.filename = filename
        db_file.global_start_time = int(filename.split(".")[0])
        db_file.aggregator = self._get_agg(path)

        self.session.commit()

    def add_agents_and_fields(self, path, overwrite=False):
        db_file = (
            self.session.query(HKFiles)
            .filter(
                HKFiles.path == path,
            )
            .one()
        )

        # line below may not be needed; is redundant
        db_agents = [a for a in db_file.agents if a.instance_id in self.iids]
        db_fields = db_file.fields

        out = self.load_fields(db_file.path, self.iids)
        fields, starts, stops, medians, means, min_vals, max_vals, stds = out

        agents = []
        for field in fields:
            agent = field.split(".")[1]
            agents.append(agent)

        #  remove duplicate agent names to avoid multiple agent_ids for
        #  same instance_id
        agents = [*set(agents)]

        # get the start and stop for each agent
        for agent in agents:
            starts_agent = []
            stops_agent = []
            for i in range(len(fields)):
                #  extract agent instance id from each field name
                if fields[i].split(".")[1] == agent:
                    #  gather all the starts and stops for each field
                    #  in the agent
                    starts_agent.append(starts[i])
                    stops_agent.append(stops[i])

            #  from the starts_agent and ends_agent, extract start
            #  and stop time for the agent
            agent_start = np.min(starts_agent)
            agent_stop = np.max(stops_agent)

            x = np.where([agent == a.instance_id for a in db_agents])[0]
            if len(x) == 0:
                # if we don't have an agent
                db_agent = HKAgents(
                    instance_id=agent,
                    start=agent_start,
                    stop=agent_stop,
                    hkfile=db_file,
                )
                self.session.add(db_agent)
            elif overwrite:
                # stop may have changed for an incomplete file?
                db_agent = db_agents[x[0]]
                db_agent.start = agent_start
                db_agent.stop = agent_stop

        self.session.commit()
        db_agents = db_file.agents

        for i in range(len(fields)):
            agentname = fields[i].split(".")[1]
            x = np.where([agentname == a.instance_id for a in db_agents])[0]
            if len(x) == 0:
                logger.error(
                    f"Somehow did not add agent {agentname} for {db_file.path}"
                )
                continue
            db_agent = db_agents[x[0]]
            x = np.where([fields[i] == f.field for f in db_fields])[0]
            if len(x) > 0 and not overwrite:
                continue
            elif len(x) == 0:
                db_field = HKFields(field=fields[i], hkfile=db_file, hkagent=db_agent)
                self.session.add(db_field)
            else:
                db_field = db_fields[x[0]]

            db_field.start = starts[i]
            db_field.stop = stops[i]
            db_field.median = medians[i]
            db_field.mean = means[i]
            db_field.min_val = min_vals[i]
            db_field.max_val = max_vals[i]
            db_field.stand_dev = stds[i]
            # update the math incase previous incomplete file?

        self.session.commit()

    def load_data(self, db_instance):
        """Load data from database objects

        Arguments
        ---------
        db_instance: HKFiles, HKAgents, HKFields, or list of one of these objects

        Returns
        -------
        load_range style dictionary. If a list of instances are passed then it
        concatenates matching keys.

        Note: the extra half second on these load_range calls is because
        load_range uses [start,stop) intervals but db_instance.stop is the last
        timestamp in the file for that field (which Katie thinks is correct for
        this specific implementation). Really, we should replace load_range with
        something that doesn't require going through time when we are really
        just asking for "all data from X in file"
        """
        if isinstance(db_instance, HKFields):
            x = hk.load_range(
                db_instance.start,
                db_instance.stop + 0.5,
                fields=[db_instance.field],
                data_dir=self.hkarchive_path,
            )
            return x
        elif isinstance(db_instance, HKAgents):
            fields = [f.field for f in db_instance.fields]
            x = hk.load_range(
                db_instance.start,
                db_instance.stop + 0.5,
                fields=fields,
                data_dir=self.hkarchive_path,
            )
            return x
        elif isinstance(db_instance, HKFiles):
            fields = [f.field for f in db_instance.fields]
            x = hk.load_range(
                min([f.start for f in db_instance.fields]),
                max([f.stop for f in db_instance.fields]) + 0.5,
                fields=fields,
                data_dir=self.hkarchive_path,
            )
            return x
        elif isinstance(db_instance, list):
            temp = [self.load_data(x) for x in db_instance]
            data = {}
            for x in temp:
                for k, item in x.items():
                    if k not in data:
                        data[k] = item
                    else:
                        data[k] = (
                            np.concatenate((data[k][0], item[0])),
                            np.concatenate((data[k][1], item[1])),
                        )
            return data
        else:
            raise ValueError(
                "db_instance must be HKFields, HKFiles, "
                "HKAgents instances or a list of these"
            )

    def get_db_agents(self, instance_id, start, stop):
        """Return list of HKAgents with given instance_id that are active
        for any subset of time between start and stop

        Arguments
        ---------
        instance_id : instance_id of agent
        start : ctime to begin search
        stop : ctime to end search

        Returns
        -------
        list of HKAgent instances
        """
        last = self.get_last_update()
        if stop > last:
            raise ValueError(
                f"stop time {stop} is beyond the last database " f"update of {last}"
            )

        agents = self.session.query(HKAgents).filter(
            db.or_(
                db.and_(HKAgents.start <= start, HKAgents.stop >= stop),
                db.and_(HKAgents.start >= start, HKAgents.start <= stop),
                db.and_(HKAgents.stop >= start, HKAgents.stop <= stop),
            )
        )
        if agents.count() == 0:
            logger.warning(f"no agents found between {start} and {stop}")

        agents = agents.filter(HKAgents.instance_id == instance_id).options(selectinload(HKAgents.fields)).all()
        return agents

    def get_last_update(self):
        """Return the last timestamp present in the database"""
        last_file = (
            self.session.query(HKFiles)
            .order_by(db.desc(HKFiles.global_start_time))
            .first()
        )
        if len(last_file.agents) == 0:
            return last_file.global_start_time
        return max([a.stop for a in last_file.agents])

    @classmethod
    def from_configs(cls, configs, iids=None):
        """
        Create a G3tHK instance from a configs dictionary

        Args
        ----
        configs - dictionary containing `data_prefix` and `g3thk_db` keys
        """
        if type(configs) == str:
            configs = load_configs(configs)

        if iids is None:
            iids = []
            if "finalization" in configs:
                servers = configs["finalization"].get("servers", {})
                for server in servers:
                    for key in server.keys():
                        # Append the value (iid) to the iids list
                        iids.append(server[key])
            else:
                logger.debug(
                    "No finalization information in configuration, agents and "
                    "fields will not be added."
                )

        return cls(
            hkarchive_path = os.path.join(configs["data_prefix"], "hk"), 
            db_path = configs["g3thk_db"],
            iids = iids
        )

    def batch_delete_files(self, file_list, dry_run=False, my_logger=None):
        for f in file_list:
            self.delete_file(
                f, dry_run=dry_run, my_logger=my_logger, commit=False
            )
        self.session.commit()

    def delete_file(self, hkfile, dry_run=False, my_logger=None, commit=True):
        """WARNING: Removes actual files from file system.

        Delete an hkfile instance, its on-disk file, and all associated agents
        and field entries in the database

        Arguments
        ---------
        hkfile: HKFile instance. Assumes file was queried uses self.session
        """

        if my_logger is None:
            my_logger = logger

        # remove field info
        my_logger.debug(f"removing field entries for {hkfile.path} from database")
        if not dry_run:
            for f in hkfile.fields:
                self.session.delete(f)

        # remove agent info
        my_logger.debug(f"removing agent entries for {hkfile.path} from database")
        if not dry_run:
            for a in hkfile.agents:
                self.session.delete(a)

        if not os.path.exists(hkfile.path):
            my_logger.warning(f"{hkfile.path} appears already deleted")
        else:
            my_logger.info(f"deleting {hkfile.path}")
            if not dry_run:
                os.remove(hkfile.path)

                base, _ = os.path.split(hkfile.path)
                if len(os.listdir(base)) == 0:
                    os.rmdir(base)

        my_logger.info(f"remove {hkfile.path} from database")
        if not dry_run:
            self.session.delete(hkfile)
            if commit:
                self.session.commit()



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simultaneous filtering and binned mapmaking.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops
from toast.observation import default_values as defaults

from .. import ops as so_ops
from .job import workflow_timer


def setup_mapmaker_ml(operators):
    """Add commandline args and operators for the S.O. maximum likelihood mapmaker.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.MLMapmaker(name="mlmapmaker", enabled=False, comps="TQU"))


@workflow_timer
def mapmaker_ml(job, otherargs, runargs, data):
    """Run the S.O. maximum likelihood mapmaker.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.mlmapmaker.enabled:
        if data.comm.group_size != 1:
            msg = "The ML mapmaker requires the process group"
            msg += " size to be exactly one, since it uses"
            msg += " threads for parallelism."
            raise RuntimeError(msg)
        job_ops.mlmapmaker.out_dir = otherargs.out_dir
        job_ops.mlmapmaker.apply(data)




# Copyright (c) 2018-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

from astropy import constants
from astropy import units as u
import numpy as np

def tb2s(tb, nu):
    """Convert blackbody temperature to spectral radiance s_nu at frequency nu.

    Args:
        tb (Quantity): Blackbody temperature
        nu (Quantity): frequency where the spectral radiance is evaluated

    Returns:
        (Quantity) s_nu with same dimensions as tb
            nominal unit: W*sr−1*m−2*Hz−1
    """
    h = constants.h
    c = constants.c
    k_b = constants.k_B
    nu = nu.to(u.Hz)
    tb = tb.to(u.K)
    x = h * nu / (k_b * tb)
    return 2 * h * nu**3 / c**2 / (np.exp(x) - 1)


def s2tcmb(s_nu, nu):
    """Convert spectral radiance s_nu at frequency nu to t_cmb.

    t_cmb is defined in the CMB community as the offset from the
    mean CMB temperature assuming a linear relation between t_cmb
    and s_nu, the t_cmb/s_nu slope is evalutated at the mean CMB
    temperature.

    Args:
        s_nu (Quantity):  spectral radiance s_nu (nominal unit: W*sr−1*m−2*Hz−1)
        nu (Quantity): frequency where the evaluation is perfomed

    Returns:
        (Quantity) t_cmb with same dimensions as tb.

    """
    T_cmb = 2.72548 * u.K  # K from Fixsen, 2009, ApJ 707 (2): 916–920
    h = constants.h
    c = constants.c
    k_b = constants.k_B
    nu = nu.to(u.Hz)
    x = h * nu / (k_b * T_cmb)

    slope = 2 * k_b * nu**2 / c**2 * ((x.value / 2) / np.sinh(x.value / 2)) ** 2
    return s_nu / slope


def tb2tcmb(tb, nu):
    """Convert blackbody temperature to t_cmb as defined above.

    Args:
        tb (Quantity): Blackbody temperature
        nu (Quantity): frequency where the spectral radiance is evaluated

    Returns:
        (Quantity) t_cmb with same dimensions as tb

    """
    s_nu = tb2s(tb, nu)
    result = s2tcmb(s_nu, nu)
    return result.decompose()



import unittest
import tempfile

from sotodlib.io import hkdb

import os
import so3g
import numpy as np
import time
from spt3g import core


def write_hk(filepath):
    """
    Writes an HK g3 file with data from an agent `test_agent`, feed `test_feed`,
    and field `test_field`.

    Returns
    ---------
    t0: float
        start time of generated data.
    t1: float
        End time of generated data.
    nsamps: int
        Number of samples in timestream
    """
    hksess = so3g.hk.HKSessionHelper()

    writer = core.G3Writer(filepath)

    frames = []
    frames.append(hksess.session_frame())
    frames.append(hksess.status_frame())

    t0 = time.time()
    t1 = t0 + 10 * 60
    sample_rate = 10 # Hz
    frame_len = 60
    times = np.arange(t0, t1, 1/sample_rate)
    data = np.zeros(len(times), dtype=float)
    frame_idxs = (times - t0) // frame_len

    prov1 = {
        'addr': 'site.test_agent.feeds.test_feed',
        'session_id': int(time.time()),
    }
    prov2 = {
        'addr': 'site.weird_agent.feeds.misc_feed',
        'session_id': int(time.time() + 5)
    }

    def hk_frame(prov, times, data, block_name='test_block'):
        frame = hksess.data_frame(0)
        tsmap = core.G3TimesampleMap()
        tsmap.times = core.G3VectorTime(times * core.G3Units.s)
        for k, v in data.items():
            tsmap[k] = core.G3VectorDouble(v)
        frame['block_names'] = core.G3VectorString([block_name])
        frame['blocks'].append(tsmap)
        frame['address'] = core.G3String(prov['addr'])
        frame['provider_session_id'] = core.G3Int(prov['session_id'])
        return frame

    for fr_idx in np.unique(frame_idxs):
        m = frame_idxs == fr_idx
        frames.append(hk_frame(prov1, times[m], {'test_field': data[m]}))
        if fr_idx % 2 == 0:
            frames.append(hk_frame(prov2, times[m], {'field1': data[m]},
                                   block_name='block1'))
        else:
            frames.append(hk_frame(prov2, times[m], {'field2': data[m]},
                                   block_name='block2'))

    for f in frames:
        writer(f)
    writer(core.G3Frame(core.G3FrameType.EndProcessing))
    return t0, t1, len(times)


class HkDbTest(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self.tempdir.cleanup()

    def test_hkdb(self):
        hk_root = os.path.join(self.tempdir.name, 'hk')
        hkdb_path = os.path.join(self.tempdir.name, 'hk.db')
        hk_file = os.path.join(hk_root, '12345_hk_test', 'hk.g3')
        os.makedirs(os.path.dirname(hk_file), exist_ok=True)
        hkcfg = hkdb.HkConfig(
            hk_root=hk_root,
            db_file=hkdb_path,
            echo_db=True,
            aliases={
                'test': 'test_agent.test_feed.test_field'
            }
        )
        t0, t1, nsamp = write_hk(hk_file)
        hkdb.update_index_all(hkcfg)
        load_spec = hkdb.LoadSpec(
            cfg=hkcfg,
            fields=['test'],
            start=t0-1, end=t1+1,
        )
        res = hkdb.load_hk(load_spec)

        self.assertEqual(len(res.test[0]), nsamp)

        feeds = hkdb.get_feed_list(load_spec)
        self.assertEqual(list(map(str, feeds)),
                         ['test_agent.test_feed.*', 'weird_agent.misc_feed.*'])

        fields = hkdb.get_field_list(load_spec)
        self.assertEqual(fields, ['test_agent.test_feed.test_field'])

        fields = hkdb.get_field_list(load_spec, fields=['weird_agent.*.*'])
        self.assertEqual(fields, ['weird_agent.misc_feed.field1',
                                  'weird_agent.misc_feed.field2'])



# Copyright (c) 2018-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import h5py
import os

import traitlets
import numpy as np
from astropy import units as u
import ephem
from scipy.constants import h, c, k
from scipy.interpolate import RectBivariateSpline
from scipy.signal import fftconvolve

import toast
from toast.timing import function_timer, Timer
from toast import qarray as qa
from toast.data import Data
from toast.traits import trait_docs, Int, Unicode, Bool, Quantity, Float, Instance
from toast.ops.operator import Operator
from toast.utils import Environment, Logger
from toast.observation import default_values as defaults
from toast.coordinates import azel_to_radec

from . import utils


def to_JD(t):
    # Unix time stamp to Julian date
    # (days since -4712-01-01 12:00:00 UTC)
    return t / 86400.0 + 2440587.5


def to_MJD(t):
    # Convert Unix time stamp to modified Julian date
    # (days since 1858-11-17 00:00:00 UTC)
    return to_JD(t) - 2400000.5


def to_DJD(t):
    # Convert Unix time stamp to Dublin Julian date
    # (days since 1899-12-31 12:00:00)
    # This is the time format used by PyEphem
    return to_JD(t) - 2415020


@trait_docs
class SimSSO(Operator):
    """Operator that generates Solar System Object timestreams."""

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    times = Unicode(
        defaults.times,
        help="Observation shared key for timestamps",
    )

    sso_name = Unicode(
        None,
        allow_none=True,
        help="Name of the SSO(s), must be recognized by pyEphem",
    )

    beam_file = Unicode(
        None,
        allow_none=True,
        help="HDF5 file that stores the simulated beam",
    )

    det_data = Unicode(
        defaults.det_data,
        help="Observation detdata key for simulated signal",
    )

    detector_pointing = Instance(
        klass=Operator,
        allow_none=True,
        help="Operator that translates boresight Az/El pointing into detector frame",
    )

    detector_weights = Instance(
        klass=Operator,
        allow_none=True,
        help="Operator that translates boresight Az/El pointing into detector weights",
    )

    finite_sso_radius = Bool(
        False, help="Treat sources as finite and convolve beam with a disc."
    )

    polarization_fraction = Float(
        0, help="Polarization fraction for all simulated SSOs",
    )

    polarization_angle = Quantity(
        0 * u.deg,
        help="Polarization angle for all simulated SSOs. Measured in the same "
        "as `detector_weights`",
    )

    @traitlets.validate("sso_name")
    def _check_sso_name(self, proposal):
        sso_names = proposal["value"]
        if sso_names is not None:
            try:
                for sso_name in sso_names.split(","):
                    sso = getattr(ephem, sso_name)()
            except AttributeError:
                raise traitlets.TraitError(f"{sso_name} is not a valid SSO name")
        return sso_names

    @traitlets.validate("beam_file")
    def _check_beam_file(self, proposal):
        beam_file = proposal["value"]
        if beam_file is not None and not os.path.isfile(beam_file):
            raise traitlets.TraitError(f"{beam_file} is not a valid beam file")
        return beam_file

    @traitlets.validate("detector_pointing")
    def _check_detector_pointing(self, proposal):
        detpointing = proposal["value"]
        if detpointing is not None:
            if not isinstance(detpointing, Operator):
                raise traitlets.TraitError(
                    "detector_pointing should be an Operator instance"
                )
            # Check that this operator has the traits we expect
            for trt in [
                "view",
                "boresight",
                "shared_flags",
                "shared_flag_mask",
                "quats",
                "coord_in",
                "coord_out",
            ]:
                if not detpointing.has_trait(trt):
                    msg = f"detector_pointing operator should have a '{trt}' trait"
                    raise traitlets.TraitError(msg)
        return detpointing

    @traitlets.validate("detector_weights")
    def _check_detector_weights(self, proposal):
        detweights = proposal["value"]
        if detweights is not None:
            if not isinstance(detweights, Operator):
                raise traitlets.TraitError(
                    "detector_weights should be an Operator instance"
                )
            # Check that this operator has the traits we expect
            for trt in [
                "view",
                "weights",
                "mode",
            ]:
                if not detweights.has_trait(trt):
                    msg = f"detector_weights operator should have a '{trt}' trait"
                    raise traitlets.TraitError(msg)
        return detweights

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Store of per-detector beam properties.  Eventually we could modify the
        # operator traits to list files per detector, per wafer, per tube, etc.
        # For now, we use the same beam for all detectors, so this will have only
        # one entry.
        self.beam_props = dict()

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()
        comm = data.comm

        for trait in "sso_name", "beam_file", "detector_pointing":
            value = getattr(self, trait)
            if value is None:
                raise RuntimeError(f"You must set `{trait}` before running SimSSO")

        for sso_name in self.sso_name.split(","):
            sso = getattr(ephem, sso_name)()

            timer = Timer()
            timer.start()

            if data.comm.group_rank == 0:
                log.debug(f"{data.comm.group} : Simulating {sso_name}")

            for obs in data.obs:
                observer = ephem.Observer()
                site = obs.telescope.site
                observer.lon = site.earthloc.lon.to_value(u.radian)
                observer.lat = site.earthloc.lat.to_value(u.radian)
                observer.elevation = site.earthloc.height.to_value(u.meter)
                observer.epoch = ephem.J2000
                observer.temp = 0  # in Celcius
                observer.compute_pressure()

                prefix = f"{comm.group} : {obs.name}"

                # Get the observation time span and compute the horizontal
                # position of the SSO
                times = obs.shared[self.times].data
                sso_az, sso_el, sso_dist, sso_diameter = self._get_sso_position(
                    data, sso, times, observer
                )
                if data.comm.group_rank == 0:
                    log.debug(
                        f"{data.comm.group} : {sso_name} at mean "
                        f"Az {np.mean(sso_az.to_value(u.degree))} deg, "
                        f"El {np.mean(sso_el.to_value(u.degree))} deg"
                    )

                # Store the SSO location

                source_coord_azel = np.column_stack(
                    [-sso_az.to_value(u.degree), sso_el.to_value(u.degree)]
                )

                obs.shared.create_column("source", (len(sso_az), 2), dtype=np.float64)
                if obs.comm.group_rank == 0:
                    obs.shared["source"].set(source_coord_azel)
                else:
                    obs.shared["source"].set(None)

                # Make sure detector data output exists.  If not, create it
                # with units of Kelvin.

                dets = obs.select_local_detectors(detectors)

                exists = obs.detdata.ensure(
                    self.det_data, detectors=dets, create_units=u.K
                )

                det_units = obs.detdata[self.det_data].units

                scale = toast.utils.unit_conversion(u.K, det_units)

                self._observe_sso(
                    data,
                    obs,
                    sso_name,
                    sso_az,
                    sso_el,
                    sso_dist,
                    sso_diameter,
                    prefix,
                    dets,
                    scale,
                )

            if data.comm.group_rank == 0:
                timer.stop()
                log.debug(
                    f"{data.comm.group} : Simulated and observed {sso_name} in "
                    f"{timer.seconds():.1f} seconds"
                )

        return

    @function_timer
    def _get_sso_temperature(self, sso_name):
        """
        Get the thermodynamic SSO temperature given
        the frequency
        """
        log = Logger.get()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        hf = h5py.File(os.path.join(dir_path, "data/planet_data.h5"), "r")
        if sso_name == "Moon":
            freq = np.linspace(0, 1000, 1001)[1:] * 1e9
            T = 300 # Kelvin
            emissivity = 1.0
            tb = 1 / (k / (h * freq) * np.log(1 + (np.exp(h * freq / (k * T)) - 1) / emissivity))
            freq = freq * 1e-9 * u.GHz
            temp = utils.tb2tcmb(tb * u.K, freq)
        elif sso_name in hf.keys():
            tb = np.array(hf.get(sso_name)) * u.K
            freq = np.array(hf.get("freqs_ghz")) * u.GHz
            temp = utils.tb2tcmb(tb, freq)
        else:
            raise ValueError(f"Unknown planet name: '{sso_name}' not in {hf.keys()}")
        return freq, temp

    @function_timer
    def _get_beam_map(self, det, sso_diameter, ttemp_det):
        """
        Construct a 2-dimensional interpolator for the beam
        """
        # Read in the simulated beam.  We could add operator traits to
        # specify whether to load different beams based on detector,
        # wafer, tube, etc and check that key here.
        log = Logger.get()
        if "ALL" in self.beam_props:
            # We have already read the single beam file.
            beam_dic = self.beam_props["ALL"]
        else:
            with h5py.File(self.beam_file, 'r') as f_t:
                beam_dic = {}
                beam_dic["data"] = f_t["beam"][:]
                beam_dic["size"] = [[f_t["beam"].attrs["size"], f_t["beam"].attrs["res"]], [f_t["beam"].attrs["npix"], 1]]
                self.beam_props["ALL"] = beam_dic
        description = beam_dic["size"]  # 2d array [[size, res], [n, 1]]
        model = beam_dic["data"]
        res = description[0][1] * u.degree
        beam_solid_angle = np.sum(model) * res**2

        n = int(description[1][0])
        size = description[0][0] * u.degree
        sso_radius_avg = np.average(sso_diameter) / 2
        sso_solid_angle = np.pi * sso_radius_avg**2
        amp = ttemp_det.to_value(u.K) * (
                    sso_solid_angle.to_value(u.rad**2) / beam_solid_angle.to_value(u.rad**2)
        )
        w = size.to_value(u.rad) / 2
        if self.finite_sso_radius:
            # Convolve the beam model with a disc rather than point-like source
            w_sso = sso_radius_avg.to_value(u.rad)
            n_sso = int(w_sso // res.to_value(u.rad)) * 2 + 3
            w_sso = (n_sso - 1) // 2 * res.to_value(u.rad)
            x_sso = np.linspace(-w_sso, w_sso, n_sso)
            y_sso = np.linspace(-w_sso, w_sso, n_sso)
            X, Y = np.meshgrid(x_sso, y_sso)
            source = np.zeros([n_sso, n_sso])
            source[X**2 + Y**2 < sso_radius_avg.to_value(u.rad)**2] = 1
            source *= amp / np.sum(source)
            model = fftconvolve(source, model, mode="full")
            # the convolved model is now larger than the pure beam model
            w += w_sso
            n += n_sso - 1
        else:
            # Treat the source as point-like. Reasonable approximation
            # if SSO radius << FWHM
            if sso_solid_angle > 0.1 * beam_solid_angle:
                log.warning("Ignoring non-negligible source diameter.  SSO image will be too narrow.")
            model *= amp
        x = np.linspace(-w, w, n)
        y = np.linspace(-w, w, n)
        beam = RectBivariateSpline(x, y, model)
        r = np.sqrt(w**2 + w**2)
        return beam, r

    @function_timer
    def _get_sso_position(self, data, sso, times, observer):
        """
        Calculate the SSO horizontal position
        """
        log = Logger.get()
        timer = Timer()
        timer.start()
        sso_az = np.zeros(times.size)
        sso_el = np.zeros(times.size)
        sso_dist = np.zeros(times.size)
        sso_diameter = np.zeros(times.size)
        for i, t in enumerate(times):
            observer.date = to_DJD(t)
            sso.compute(observer)
            sso_az[i] = sso.az
            sso_el[i] = sso.alt
            sso_dist[i] = sso.earth_distance
            sso_diameter[i] = sso.size
        sso_az *= u.radian
        sso_el *= u.radian
        sso_dist *= u.AU
        sso_diameter *= u.arcsec
        if data.comm.group_rank == 0:
            timer.stop()
            log.verbose(
                f"{data.comm.group} : Computed {sso.name} position in "
                f"{timer.seconds():.1f} seconds"
            )
        return sso_az, sso_el, sso_dist, sso_diameter

    @function_timer
    def _observe_sso(
        self,
        data,
        obs,
        sso_name,
        sso_az,
        sso_el,
        sso_dist,
        sso_diameter,
        prefix,
        dets,
        scale,
    ):
        """
        Observe the SSO with each detector in tod
        """
        log = Logger.get()
        timer = Timer()

        if self.polarization_fraction != 0 and self.detector_weights is None:
            raise RuntimeError(
                "Cannot simulate polarized SSOs without detector weights"
            )

        # Get a view of the data which contains just this single
        # observation
        obs_data = data.select(obs_name=obs.name)

        zaxis = np.array([0.0, 0.0, 1.0])
        bore_quat = obs_data.obs[0].shared[defaults.boresight_azel][:]
        bore_lon, bore_lat, _ = qa.to_lonlat_angles(bore_quat)

        beam = None
        for idet, det in enumerate(dets):
            timer.clear()
            timer.start()
            bandpass = obs.telescope.focalplane.bandpass
            signal = obs.detdata[self.det_data][det]

            self.detector_pointing.apply(obs_data, detectors=[det])
            if self.polarization_fraction != 0:
                self.detector_weights.apply(obs_data, detectors=[det])
            det_quat = obs_data.obs[0].detdata[self.detector_pointing.quats][det]

            # Convert Az/El quaternion of the detector into angles
            theta, phi, _ = qa.to_iso_angles(det_quat)

            # Azimuth is measured in the opposite direction
            # than longitude
            az = -phi
            el = np.pi / 2 - theta

            mean_az = np.mean(az)
            mean_el = np.mean(el)

            # Convolve the planet SED with the detector bandpass
            sso_freq, sso_temp = self._get_sso_temperature(sso_name)
            det_temp = bandpass.convolve(det, sso_freq, sso_temp) * u.K
            if beam is None or not "ALL" in self.beam_props:
                beam, radius = self._get_beam_map(det, sso_diameter, det_temp)

            # Interpolate the beam map at appropriate locations

            az_diff = (az - sso_az.to_value(u.rad) + np.pi) % (2 * np.pi) - np.pi
            x = az_diff * np.cos(el)
            y = el - sso_el.to_value(u.rad)
            r = np.sqrt(x**2 + y**2)

            good = r < radius
            sig = beam(x[good], y[good], grid=False)
            if self.polarization_fraction != 0:
                self._observe_polarization(sig, obs, det, good)
            signal[good] += scale * sig

            timer.stop()
            if data.comm.world_rank == 0:
                log.verbose(
                    f"{prefix} : Simulated and observed {sso_name} in {det} in "
                    f"{timer.seconds():.1f} seconds"
                )

        return

    def _observe_polarization(self, sig, obs, det, good):
        # Stokes weights for observing polarized source
        weights = obs.detdata[self.detector_weights.weights][det]
        weight_mode = self.detector_weights.mode
        if "I" in weight_mode:
            ind = weight_mode.index("I")
            weights_I = weights[good, ind].copy()
        else:
            weights_I = 0
        if "Q" in weight_mode:
            ind = weight_mode.index("Q")
            weights_Q = weights[good, ind].copy()
        else:
            weights_Q = 0
        if "U" in weight_mode:
            ind = weight_mode.index("U")
            weights_U = weights[good, ind].copy()
        else:
            weights_U = 0

        pfrac = self.polarization_fraction
        angle = self.polarization_angle.to_value(u.radian)

        sig *= weights_I + pfrac * (
            np.cos(2 * angle) * weights_Q + np.sin(2 * angle) * weights_U
        )
        return


    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "shared": [
                self.times,
            ],
        }
        return req

    def _provides(self):
        return {
            "detdata": [
                self.det_data,
            ]
        }

    def _accelerators(self):
        return list()


# def plot_projected_quats(
#     outfile, qbore=None, qdet=None, valid=slice(None), scale=1.0, planet=None
# ):
#     """Plot a list of quaternion arrays in longitude / latitude."""

#     toast.vis.set_matplotlib_backend()
#     import matplotlib.pyplot as plt

#     # Convert boresight and detector quaternions to angles

#     qbang = None
#     if qbore is not None:
#         qbang = np.zeros((3, qbore.shape[0]), dtype=np.float64)
#         qbang[0], qbang[1], qbang[2] = qa.to_lonlat_angles(qbore)
#         qbang[0] *= 180.0 / np.pi
#         qbang[1] *= 180.0 / np.pi
#         lon_min = np.amin(qbang[0])
#         lon_max = np.amax(qbang[0])
#         lat_min = np.amin(qbang[1])
#         lat_max = np.amax(qbang[1])

#     qdang = None
#     if qdet is not None:
#         qdang = np.zeros((qdet.shape[0], 3, qdet.shape[1]), dtype=np.float64)
#         for det in range(qdet.shape[0]):
#             qdang[det, 0], qdang[det, 1], qdang[det, 2] = qa.to_lonlat_angles(qdet[det])
#             qdang[det, 0] *= 180.0 / np.pi
#             qdang[det, 1] *= 180.0 / np.pi
#         lon_min = np.amin(qdang[:, 0])
#         lon_max = np.amax(qdang[:, 0])
#         lat_min = np.amin(qdang[:, 1])
#         lat_max = np.amax(qdang[:, 1])

#     # Set the sizes of shapes based on the plot range

#     span_lon = lon_max - lon_min
#     span_lat = lat_max - lat_min
#     span = max(span_lon, span_lat)
#     # bmag = 0.5 * span * scale
#     # dmag = 0.2 * span * scale
#     bmag = 0.2
#     dmag = 0.1

#     if span_lat > span_lon:
#         fig_y = 10
#         fig_x = fig_y * (span_lon / span_lat)
#         if fig_x < 4:
#             fig_x = 4
#     else:
#         fig_x = 10
#         fig_y = fig_x * (span_lat / span_lon)
#         if fig_y < 4:
#             fig_y = 4

#     figdpi = 100

#     fig = plt.figure(figsize=(fig_x, fig_y), dpi=figdpi)
#     ax = fig.add_subplot(1, 1, 1, aspect="equal")

#     # Compute the font size to use for detector labels
#     fontpix = 0.1 * figdpi
#     fontpt = int(0.75 * fontpix)

#     # Plot source if we have it

#     if planet is not None:
#         ax.plot(planet[:, 0], planet[:, 1], color="purple", marker="+")

#     # Plot boresight if we have it

#     if qbang is not None:
#         ax.scatter(qbang[0][valid], qbang[1][valid], color="black", marker="x")
#         for ln, lt, ps in np.transpose(qbang)[valid]:
#             wd = 0.05 * bmag
#             dx = bmag * np.sin(ps)
#             dy = -bmag * np.cos(ps)
#             ax.arrow(
#                 ln,
#                 lt,
#                 dx,
#                 dy,
#                 width=wd,
#                 head_width=4.0 * wd,
#                 head_length=0.2 * bmag,
#                 length_includes_head=True,
#                 ec="red",
#                 fc="red",
#             )

#     # Plot detectors if we have them

#     if qdang is not None:
#         for idet, dang in enumerate(qdang):
#             ax.scatter(dang[0][valid], dang[1][valid], color="blue", marker=".")
#             for ln, lt, ps in np.transpose(dang)[valid]:
#                 wd = 0.05 * dmag
#                 dx = dmag * np.sin(ps)
#                 dy = -dmag * np.cos(ps)
#                 ax.arrow(
#                     ln,
#                     lt,
#                     dx,
#                     dy,
#                     width=wd,
#                     head_width=4.0 * wd,
#                     head_length=0.2 * dmag,
#                     length_includes_head=True,
#                     ec="blue",
#                     fc="blue",
#                 )
#             ax.text(
#                 dang[0][valid][0] + (idet % 2) * 1.5 * dmag,
#                 dang[1][valid][0] + 1.0 * dmag,
#                 f"{idet:02d}",
#                 color="k",
#                 fontsize=fontpt,
#                 horizontalalignment="center",
#                 verticalalignment="center",
#                 bbox=dict(fc="w", ec="none", pad=1, alpha=0.0),
#             )

#     # Invert x axis so that longitude reflects what we would see from
#     # inside the celestial sphere
#     plt.gca().invert_xaxis()

#     ax.set_xlabel("Longitude Degrees", fontsize="medium")
#     ax.set_ylabel("Latitude Degrees", fontsize="medium")

#     fig.suptitle("Projected Pointing and Polarization on Sky")

#     plt.savefig(outfile)
#     plt.close()



# Copyright (c) 2020-2022 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simons Observatory TOAST I/O helper classes.

"""

from .save_book import (
    book_name_from_obs,
    write_book,
)

from .load_book import (
    parse_book_name,
    parse_book_time,
    read_book,
)



import numpy as np
from functools import reduce
from collections import OrderedDict as odict

from so3g.proj import Ranges, RangesMatrix
from . import AxisManager

try:
    from scipy.sparse import csr_array
except ImportError:
    from scipy.sparse import csr_matrix as csr_array

class FlagManager(AxisManager):
    """An extension of the AxisManager class to make functions 
    more specifically associated with cuts and flags.
    
    FlagManagers must have a dets axis and a samps axis when created.
    
    FlagManager only expects to have individual flags that are mapped to the 
    detector axis, the sample axis, or both. 
    
    Detector Flags can be passed as bitmasks or boolean arrays. To match with
    Ranges and RangesMatrix, the default is False and the exceptions
    are True
    """
    
    def __init__(self, dets_axis, samps_axis):
        self._dets_name = dets_axis.name
        self._samps_name = samps_axis.name
        
        super().__init__(dets_axis, samps_axis)
                
    
    def wrap(self, name, data, axis_map=None, **kwargs):
        """See core.AxisManager for basic usage
        
        If axis_map is None, the data better be (dets,), (samps,),
            or (dets, samps). Will not work if dets.count == samps.count
            
        """
        
        if axis_map is None:
            if self[self._dets_name].count == self[self._samps_name].count:
                raise ValueError("Cannot auto-detect axis_map when dets and "
                                 "samps axes have equal lengths. axis_map "
                                 "must be defined")
            s = _get_shape(data)
            
            if len(s) == 1:
                if s[0] == self[self._dets_name].count:
                    ## detector only flag. Turn into RangesMatrix
                    axis_map=[(0,self._dets_name)]
                elif s[0] == self[self._samps_name].count:
                    axis_map=[(0, self.samps)]
                else:
                    raise ValueError("FlagManager only takes data aligned with"
                                     " dets and/or samps. Data of shape {}"
                                     " is the wrong shape".format(s))
            elif len(s) == 2:
                if s[0] == self[self._dets_name].count and s[1] == self[self._samps_name].count:
                    axis_map=[(0,self._dets_name), (1,self._samps_name)]
                elif s[1] == self[self._dets_name].count and s[0] == self[self._samps_name].count:
                    raise ValueError("FlagManager only takes 2D data aligned as"
                                     " (dets, samps). Data of shape {}"
                                     " is the wrong shape".format(s))
                else:
                    raise ValueError("FlagManager only takes 2D data aligned as"
                                     " (dets, samps). Data of shape {}"
                                     " is the wrong shape".format(s))
            else:
                raise ValueError("FlagManager only takes data aligned with"
                                     " dets and/or samps. Data of shape {}"
                                     " is the wrong shape".format(s))
        
        if len(axis_map)==1 and axis_map[0][1]==self._dets_name:
            ### Change detector flags to RangesMatrix in the backend
            x = Ranges(self.samps.count)
            data = RangesMatrix([Ranges.ones_like(x) if Y 
                                 else Ranges.zeros_like(x) for Y in data])
            axis_map = [(0,self._dets_name),(1,self._samps_name)]

        return super().wrap(name, data, axis_map, **kwargs)

    def wrap_dets(self, name, data):
        """Adding flag with just (dets,) axis.
        """
        s = _get_shape(data)
        if not len(s) == 1 or s[0] != self[self._dets_name].count:
            raise ValueError("Data of shape {} is cannot be aligned with"
                             "the detector axis".format(s))
        return self.wrap(name, data, axis_map=[(0,self._dets_name)])
        
    def wrap_samps(self, name, data):
        """Adding flag with just (samps,) axis.
        """
        s = _get_shape(data)
        if not len(s) == 1 or s[0] != self[self._samps_name].count:
            raise ValueError("Data of shape {} is cannot be aligned with"
                             "the samps axis".format(s))
        return self.wrap(name, data, axis_map=[(0,self._samps_name)])
        
    def wrap_dets_samps(self, name, data):
        """Adding flag with (dets, samps) axes.
        """
        s = _get_shape(data)
        if (not len(s) == 2 or s[0] != self[self._dets_name].count or
               s[1] != self[self._samps_name].count):
            raise ValueError("Data of shape {} is cannot be aligned with"
                             "the (dets,samps) axss".format(s))
        return self.wrap(name, data, axis_map=[(0,self._dets_name), (1,self._samps_name)])

        
    def copy(self, axes_only=False):
        out = FlagManager(self[self._dets_name], self[self._samps_name])
        for k, v in self._axes.items():
            out._axes[k] = v
        if axes_only:
            return out
        for k, v in self._fields.items():
            out._fields[k] = v.copy()
        for k, v in self._assignments.items():
            out._assignments[k] = v.copy()
        return out
    
    def get_zeros(self, wrap=None):
        """
        Return a correctly sized RangesMatrix for building cuts for the FlagManager
        
        Args:
            wrap: if not None, it is a string with which to add to the FlagManager
        """
        out = RangesMatrix([Ranges(self[self._samps_name].count) for det in self[self._dets_name].vals])
        if not wrap is None:
            self.wrap_dets_samps( wrap, out)
            return self[wrap]
        return out
        
    def buffer(self, n_buffer, flags=None):
        """Buffer all the samps cuts by n_buffer
        Like with Ranges / Ranges Matrix, buffer changes everything in place        
        
        Args:
            n_buffer: number of samples to buffer the samps cuts
            flags: List of flags to buffer. Uses their names
        """
        if flags is None:
            flags = self._fields
        
        for f in flags:
            self[f].buffer(n_buffer)
        
    def buffered(self, n_buffer, flags=None):
        """Return new FlagManager that has all the samps cuts buffered by n_buffer
        Like with Ranges / Ranges Matrix, buffered returns new object
        
        Args:
            n_buffer: number of samples to buffer the samps cuts
            flags: List of flags to buffer. Uses their names
        
        Returns:
            new: FlagManager with all flags buffered
        """
        new = self.copy()
        new.buffer(n_buffer, flags)
        return new
        
    def reduce(self, flags=None, method='union', wrap=False, new_flag=None,
               remove_reduced=False):
        """Reduce (combine) flags in the FlagManager together. 
        
        Args:
            flags: List of flags to collapse together. Uses their names.
                   If flags is None then all flags are reduced
            method: How to collapse the data. Accepts 'union','intersect',
                        or function.
            wrap: if True, add reduced flag to self
            new_flag: name of new flag, required if wrap is True
            remove_reduced: if True, remove all reduced flags from self
        
        Returns:
            out: reduced flag
        """
        if flags is None:
            ## copy needed to no break things if removing flags
            flags = self._fields.copy()

        to_reduce = [self._fields[f] for f in flags]
        if len(flags)==0:
            raise ValueError('Found zero flags to combine')
            
        out = self.get_zeros()
        
        ## need to add out to prevent flag ordering from causing errors
        ### (Ranges can't add to RangeMatrix, only other way around)
        to_reduce[0] = out+to_reduce[0]
        
        if method == 'union': 
            op = lambda x, y: x+y
        elif method == 'intersect':
            op = lambda x, y: x*y
        else:
            op = method
        out = reduce(op, to_reduce)
        
        # drop the fields if needed
        if remove_reduced: 
            for f in flags: 
                self.move(f, None)
                
        if wrap:
            if new_flag is None:
                raise ValueError("new_flag cannot be None if wrap is True")
            self.wrap(new_flag, out)
            
        return out        
    
    def has_cuts(self, flags=None):
        """Return list of detector ids that have cuts
        
        Args: 
            flags: [optional] If not none it is the list of flags to combine to see
                    if cuts exist
        """
        c = self.reduce(flags=flags)
        idx = [len(x.ranges())>0 for x in c]
        return self[self._dets_name].vals[idx]

    @classmethod
    def for_tod(cls, tod, dets_name='dets', samps_name='samps'):
        """Create a Flag manager for an AxisManager tod which has axes for detectors
        and samples.
        
        Args:
            tod: AxisManager for the specific data
            dets_name: name of the axis that should be treated as detectors
            samps_name: name of the axis that should be treated as samples
        """
        
        return cls(tod[dets_name], tod[samps_name])

    @classmethod
    def promote(cls, src, dets_name, samps_name):
        """Create an instance of this class from an AxisManager.  The axes in
        src with names dets_name and samps_name are used for the dets
        and samps axes.

        This is a move-style constructor, and empties out src to
        minimize copies.

        """
        flagman = cls(src._axes.pop(dets_name),
                      src._axes.pop(samps_name))
        flagman._axes.update(src._axes)
        flagman._fields = dict(src._fields)
        flagman._assignments = dict(src._assignments)

        # Reset source.
        src._axes = odict()
        src._assignments = {}
        src._fields = odict()

        return flagman

def _get_shape(data):
    try:
        return data.shape
    except:
        ### catches if a detector mask is just a list
        return np.shape(data)

def has_any_cuts(flag):
    return np.array([len(x.ranges())>0 for x in flag], dtype='bool')
def has_all_cut(flag):
    return np.array(
        [len(x.complement().ranges())==0 for x in flag],
        dtype='bool'
    )
def count_cuts(flag):
    return np.array([len(x.ranges()) for x in flag], dtype='int')

def has_ratio_cuts(flag, ratio):
    """Determine if the ratio of flag samples to total samples in each flag 
    exceeds a given threshold.

    Args:
      flag (RangesMatrix): so3g.proj.RangesMatrix
      ratio (float or int): The threshold ratio (0 =< ratio =< 1) above which a flag is 
        considered to have too many cuts.
    """

    if not isinstance(ratio, (float, int)) or (ratio < 0 or ratio > 1):
        raise ValueError('Ratio must be float or int and between 0 and 1')
    len_samps = flag.shape[-1]
    return np.array([np.sum(np.diff(x.ranges()))/len_samps > ratio for x in flag])

def flag_cut_select(flags, kind, invert=False):
    """Determine which detectors to select based on flag conditions.

    Args:
        flags (RangesMatrix): An instance of so3g.proj.RangesMatrix indicating flagged time ranges.
        kind (str or float): One of the following:
            - 'any': Select/cut detectors with any flagged samples.
            - 'all': Select/cut detectors with all samples flagged.
            - float: A threshold ratio (0.0–1.0); selects/cuts detectors whose flagged ratio exceeds the threshold.
        intert (bool): default=False returns detectors to be excluded =True, detectors to be kept = False. If True, The logic is flipped.

    Returns:
        np.ndarray: Boolean array indicating which detectors to **keep** (True) or **drop** (False),
                    depending on `cut` and `kind`.

    Examples:
        1. invert=False, kind='any' → Select detectors with **no** True flags (e.g., for Moon cut).
        2. invert=True, kind='any' → Select detectors with **any** True flags (e.g., for planet selection).
        3. invert=False, kind=0.4 → Select detectors with <40% of True flags.
    """
    if invert:
        if kind == 'any':
            return has_any_cuts(flags)
        elif kind == 'all':
            return has_all_cut(flags)
        elif isinstance(kind, float):
            return has_ratio_cuts(flags, ratio=kind)
        else:
            raise ValueError("kind must be 'any', 'all', or a float between 0.0 and 1.0")
    else:
        if kind == 'any':
            return ~has_any_cuts(flags)
        elif kind == 'all':
            return ~has_all_cut(flags)
        elif isinstance(kind, float):
            return ~has_ratio_cuts(flags, ratio=kind)
        else:
            raise ValueError("kind must be 'any', 'all', or a float between 0.0 and 1.0")


def sparse_to_ranges_matrix(arr, buffer=0, close_gaps=0, val=True):
    """Convert a csr sparse array into a ranges matrix
    
    Arguments:
    
    arr: sparse csr boolean array
    buffer: integer samples to buffer around Ranges
    close_gaps: any integer sample gaps to close in the Ranges
    val: what value in the boolean array indicates a flag
    """
    x = RangesMatrix.zeros(arr.shape)
    for i in range(arr.shape[0]):
        slc = arr.indices[arr.indptr[i]:arr.indptr[i+1]]
        if not np.all(arr.data[arr.indptr[i]:arr.indptr[i+1]]==val):
            raise
        for s in slc:
            x[i].append_interval_no_check(int(s),int(s+1))
        x[i].buffer(buffer)
        x[i].close_gaps(close_gaps)
    return x

def find_common_edge_idx(flags):
    """Find the common valid range across multiple RangesMatrix objects.

    Args:
        flags (RangesMatrix): An instance of so3g.proj.RangesMatrix indicating flagged time ranges.
    Returns:
        min_idx, max_idx: minmum and maximum indices that has False flag across all detectros.
    """
    max_val = max(arr.complement().ranges()[:, 1].max() for arr in flags)
    common_mask = np.ones(max_val, dtype=bool)
    for arr in flags:
        mask = np.zeros(max_val, dtype=bool)
        for start, end in arr.complement().ranges():
            mask[start:end] = True
        common_mask &= mask

    valid_indices = np.where(common_mask)[0]
    if len(valid_indices) == 0:
        raise ValueError("No common valid range found across all flags.")
    return valid_indices[0], valid_indices[-1]



# Copyright (c) 2022-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import re
import sqlite3
import yaml
from datetime import datetime, timezone

import numpy as np
from astropy import units as u
from astropy.table import Column, QTable

import toast
from toast.timing import function_timer, Timer
from toast.traits import (
    trait_docs,
    Int,
    Unicode,
    Instance,
    List,
    Unit,
    Dict,
    Bool,
    Float,
)
from toast.ops.operator import Operator
from toast.ops.pipeline import Pipeline
from toast.utils import Logger
from toast.dist import distribute_discrete
from toast.observation import default_values as defaults

import so3g

from ...core import Context, AxisManager, FlagManager
from ...core.axisman import AxisInterface

from ..instrument import SOSite

from .load_context_utils import (
    compute_boresight_pointing,
    parse_metadata,
    read_and_preprocess_wafers,
    open_context,
    distribute_detector_data,
    distribute_detector_props,
    ax_name_fp_subst,
)


@trait_docs
class LoadContext(Operator):
    """Load one or more observations from a Context into observing sessions.

    Given a context, load one or more observations.  If specified, the context
    should exist on all processes in the group.  The `readout_ids`, `detsets`,
    and `bands` traits are used to select subsets of detectors.  Alternatively
    the context file can be specified, in which case this is passed to the
    context constructor.

    The traits starting with an "ax_*" prefix indicate the axis manager names
    that should be mapped to the standard toast shared and detdata objects.
    For `ax_flags` tuples, a negative bit value indicates that the flag data
    should first be inverted before combined.

    For all of the "ax_*" traits, the names may additionally include substitution
    strings enclosed in '{}' that contain the name of a focalplane detector_data
    key.  For example, the shared smurfgap flags are named after the UFM (array
    name).  So you can import the correct field name for all observations with:

        ax_flags=[
            (
                "smurfgaps_ufm_{det_info:wafer:array}",
                defaults.shared_mask_invalid
            ),
        ]

    Additional nested axismanager data is grouped into detdata, shared, or
    metadata objects based on whether they use detector / sample axes.  The
    nested structure is flattened into names built from the keys in the
    hierarchy.

    Important considerations for data distribution:

    - Each "observation" in the Context system maps to a toast "observing
      session".  By default, a Context observation is turned into a toast
      observing session with one observation per wafer.  If `combine_wafers`
      is True, then all wafers will be combined into a single toast
      observation.

    - In the default case of one observation per wafer, recall that only
      one process in a group loads data from disk.  For jobs that are I/O
      bound and loading multiple wafers, you should use multiple process
      groups to parallelize the data loading (which is the dominant cost).

    """

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    context = Instance(
        klass=Context,
        allow_none=True,
        help="The Context, which should exist on all processes",
    )

    context_file = Unicode(
        None,
        allow_none=True,
        help="Create a context from this file",
    )

    observation_regex = Unicode(
        None,
        allow_none=True,
        help="Regular expression match to apply to observation IDs",
    )

    observation_file = Unicode(
        None,
        allow_none=True,
        help="Text file containing observation IDs to load",
    )

    preprocess_config = Unicode(
        None,
        allow_none=True,
        help="Apply site-pipeline pre-processing with this configuration",
    )

    ignore_preprocess_archive = Bool(
        False,
        help="If True alway compute preprocess on the fly, don't load from archive.",
    )

    observations = List(list(), help="List of observation IDs to load")

    readout_ids = List(list(), help="Only load this list of readout_id values")

    stream_ids = List(list(), help="Only load this list of stream_id values")

    detsets = List(list(), help="Only load this list of detset values")

    bands = List(list(), help="Only load this list of band values")

    dets_select = Dict(
        dict(), help="The `dets` selection dictionary to pass to get_obs()"
    )

    combine_wafers = Bool(
        False,
        help="If True, combine all wafers into a single observation",
    )

    ax_times = Unicode(
        "timestamps",
        help="Name of field to associate with times",
    )

    ax_flags = List(
        [],
        help="Tuples of (field, bit value) merged to shared_flags",
    )

    ax_det_signal = Unicode(
        "signal",
        allow_none=True,
        help="Name of field to associate with det_data",
    )

    ax_det_flags = List(
        [],
        help="Tuples of (field, bit_value) merged to det_flags",
    )

    ax_boresight_az = Unicode(
        "boresight:az",
        allow_none=True,
        help="Field with boresight Az",
    )

    ax_boresight_el = Unicode(
        "boresight:el",
        allow_none=True,
        help="Field with boresight El",
    )

    ax_boresight_roll = Unicode(
        "boresight:roll",
        allow_none=True,
        help="Field with boresight Roll",
    )

    ax_hwp_angle = Unicode(
        "hwp_angle",
        allow_none=True,
        help="Field with HWP angle",
    )

    ax_pathsep = Unicode(
        ":",
        help="Path separator when flattening nested fields",
    )

    ax_detinfo_wafer_key = Unicode(
        "stream_id",
        help="Name of the det_info property containing the wafer ID",
    )

    axis_detector = Unicode(
        "dets", help="Name of the LabelAxis for the detector direction"
    )

    axis_sample = Unicode(
        "samps", help="Name of the OffsetAxis for the sample direction"
    )

    telescope_name = Unicode("UNKNOWN", help="Name of the telescope")

    detset_key = Unicode(
        None,
        allow_none=True,
        help="Column of the focalplane detector_data to use for data distribution",
    )

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    det_data_units = Unit(
        defaults.det_data_units, help="Output units if creating detector data"
    )

    shared_flags = Unicode(
        defaults.shared_flags,
        allow_none=True,
        help="Observation shared key for common flags",
    )

    det_data = Unicode(
        defaults.det_data,
        allow_none=True,
        help="Observation detdata key for detector signal",
    )

    det_flags = Unicode(
        defaults.det_flags,
        allow_none=True,
        help="Observation detdata key for detector flags",
    )

    azimuth = Unicode(
        defaults.azimuth, help="Observation shared key for boresight Azimuth"
    )

    elevation = Unicode(
        defaults.elevation, help="Observation shared key for boresight Elevation"
    )

    roll = Unicode(defaults.roll, help="Observation shared key for boresight Roll")

    boresight_azel = Unicode(
        defaults.boresight_azel,
        help="Observation shared key for boresight Az/El quaternions",
    )

    boresight_radec = Unicode(
        defaults.boresight_radec,
        help="Observation shared key for boresight RA/DEC quaternions",
    )

    corotator_angle = Unicode(
        None,
        allow_none=True,
        help="Observation shared key for corotator_angle (if it is used)",
    )

    boresight_angle = Unicode(
        None,
        allow_none=True,
        help="Observation shared key for boresight rotation angle (if it is used)",
    )

    hwp_angle = Unicode(
        None,
        allow_none=True,
        help="Observation shared key for HWP angle (if it is used)",
    )

    analytic_bandpass = Bool(False, help="Add analytic bandpass to each detector")

    bandwidth = Float(0.2, help="Fractional bandwith used in analytic bandpass")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()
        timer = Timer()
        timer.start()

        comm = data.comm

        if self.combine_wafers:
            msg = "Combining multiple wafers in a single observation is"
            msg += " currently broken"
            raise NotImplementedError(msg)

        if self.context is None:
            if self.context_file is None:
                msg = "Either the context or context_file must be specified"
                raise RuntimeError(msg)
        else:
            if self.context_file is not None:
                msg = "Only one of the context or context_file should be specified"
                raise RuntimeError(msg)

        # Build our detector selection dictionary.  Merge our explicit traits
        # with any pre-existing detector selection.
        dets_select = None
        if len(self.dets_select) > 0:
            # Use the full dictionary provided by the user
            dets_select = self.dets_select
        elif (
            len(self.readout_ids) > 0
            or len(self.bands) > 0
            or len(self.detsets) > 0
            or len(self.stream_ids) > 0
        ):
            # We have some selection, build the dictionary
            dets_select = dict()
            if len(self.readout_ids) > 0:
                dets_select["readout_id"] = list(self.readout_ids)
            if len(self.bands) > 0:
                dets_select["band"] = list(self.bands)
            if len(self.detsets) > 0:
                dets_select["detset"] = list(self.detsets)
            if len(self.stream_ids) > 0:
                dets_select["stream_id"] = list(self.stream_ids)

        # Get any selection of wafers
        keep_wafers = None
        if dets_select is not None and "stream_id" in dets_select:
            if isinstance(dets_select["stream_id"], (list, tuple, set)):
                keep_wafers = set(dets_select["stream_id"])
            else:
                # Scalar
                keep_wafers = set()
                keep_wafers.add(dets_select["stream_id"])

        # One global process queries the observation metadata and computes
        # the observation distribution among process groups.
        obs_check = (
            (self.observation_regex is not None)
            + (len(self.observations) > 0)
            + (self.observation_file is not None)
        )
        if obs_check != 1:
            msg = "Exactly one of observation_regex, observation_file"
            msg += " and observations should be specified"
            raise RuntimeError(msg)
        if self.observation_file is not None:
            olist = None
            if comm.world_rank == 0:
                olist = list()
                with open(self.observation_file, "r") as f:
                    for line in f:
                        if re.match(r"^#.*", line) is None:
                            olist.append(line.strip())
            if comm.comm_world is not None:
                olist = comm.comm_world.bcast(olist, root=0)
            self._observations = olist
        else:
            self._observations = self.observations

        obs_props = None
        preproc_conf = None
        if comm.world_rank == 0:
            obs_list = list(sorted(self._observations))
            if self.preprocess_config is not None:
                with open(self.preprocess_config, "r") as f:
                    preproc_conf = yaml.safe_load(f)

            # Query the obsdb directly so that we only have one DB query,
            # rather than doing one query per observation.
            sort_by = ["obs_id"]
            if len(obs_list) > 0:
                query_str = "obs.obs_id IN ("
                for oid in obs_list:
                    query_str += f"'{oid}',"
                query_str = query_str.rstrip(",")
                query_str += ")"
            else:
                query_str = "1"

            # If we are using preprocessing, load the archive DB and cut any
            # observations or wafer slots that do not exist.
            preproc_lookup = None
            if (preproc_conf is not None) and (not self.ignore_preprocess_archive):
                preproc_lookup = dict()
                preproc_db = preproc_conf["archive"]["index"]
                con = sqlite3.connect(preproc_db)
                cur = con.cursor()
                res = cur.execute('select "obs:obs_id","dets:wafer_slot" from map')
                for row in res.fetchall():
                    obid = row[0]
                    wslot = row[1]
                    if obid not in preproc_lookup:
                        preproc_lookup[obid] = set()
                    preproc_lookup[obid].add(wslot)
                del res
                del cur
                con.close()
                del con

            # Open the databases and query
            ctx = open_context(context=self.context, context_file=self.context_file)
            query_result = ctx.obsdb.query(query_text=query_str, sort=sort_by)

            # Build up the list of observing session properties, and optionally
            # only keep observations that match a regex.
            pat = None
            if self.observation_regex is not None:
                pat = re.compile(self.observation_regex)
            session_props = list()
            for row in query_result:
                obs_id = str(row["obs_id"])
                if pat is not None and pat.match(obs_id) is None:
                    continue
                # If we are using preprocessing, only keep obs_ids and wafer slots that
                # exist in the preprocessing archive.
                raw_wafer_slots = list(row["wafer_slots_list"].split(","))
                raw_stream_ids = list(row["stream_ids_list"].split(","))
                if preproc_lookup is not None:
                    if obs_id not in preproc_lookup:
                        msg = f"Requested obs_id {obs_id} does not exist in "
                        msg += f"preprocessing archive {preproc_db}.  Skipping."
                        log.warning(msg)
                        continue
                    wafer_slots = list()
                    stream_ids = list()
                    for ws, sid in zip(raw_wafer_slots, raw_stream_ids):
                        if ws not in preproc_lookup[obs_id]:
                            msg = f"Wafer slot {obs_id}:{ws} not in preprocessing"
                            msg += f" archive {preproc_db}.  Skipping."
                            log.warning(msg)
                        else:
                            wafer_slots.append(ws)
                            stream_ids.append(sid)
                else:
                    wafer_slots = raw_wafer_slots
                    stream_ids = raw_stream_ids
                sprops = dict()
                sprops["session_name"] = obs_id
                sprops["session_start"] = float(row["start_time"])
                sprops["session_stop"] = float(row["stop_time"])
                sprops["n_samples"] = int(row["n_samples"])
                sprops["tele_name"] = str(row["telescope"])
                sprops["n_wafers"] = int(row["wafer_count"])
                sprops["wafer_slots"] = wafer_slots
                sprops["wafers"] = stream_ids
                session_props.append(sprops)

            # Close the databases
            if self.context_file is not None:
                del ctx

            # Now construct the list of observation properties, either with all wafers
            # in a single observation or one observation per wafer.
            obs_props = list()
            for sprops in session_props:
                if self.combine_wafers:
                    # Place all wafers into a single observation
                    oprops = dict(sprops)
                    oprops["name"] = oprops["session_name"]
                    oprops["wafer"] = "all"
                    if keep_wafers is not None:
                        new_wafers = list()
                        for wf in oprops["wafers"]:
                            if wf in keep_wafers:
                                new_wafers.append(wf)
                        oprops["n_wafers"] = len(new_wafers)
                        oprops["wafers"] = new_wafers
                    if oprops["n_wafers"] > 0:
                        obs_props.append(oprops)
                else:
                    # One observation per wafer
                    for wf in sprops["wafers"]:
                        if (keep_wafers is None) or (wf in keep_wafers):
                            oprops = dict(sprops)
                            oprops["n_wafers"] = 1
                            oprops["name"] = f"{oprops['session_name']}_{wf}"
                            oprops["wafer"] = wf
                            obs_props.append(oprops)

        if comm.comm_world is not None:
            obs_props = comm.comm_world.bcast(obs_props, root=0)
            if self.preprocess_config is not None:
                preproc_conf = comm.comm_world.bcast(preproc_conf, root=0)

        log.info_rank(
            "LoadContext parsed observation sizes in", comm=comm.comm_world, timer=timer
        )

        if len(obs_props) == 0:
            msg = "No observation IDs specified or matched the regex"
            raise RuntimeError(msg)

        # Distribute observations among groups- we use the detector-seconds
        # for load balancing.

        obs_sizes = [int(x["n_wafers"] * x["n_samples"]) for x in obs_props]
        groupdist = distribute_discrete(obs_sizes, comm.ngroups)
        group_firstobs = groupdist[comm.group].offset
        group_numobs = groupdist[comm.group].n_elem

        # Every group loads its observations
        for obindx in range(group_firstobs, group_firstobs + group_numobs):
            obs_name = obs_props[obindx]["name"]
            n_samp = obs_props[obindx]["n_samples"]
            otimer = Timer()
            otimer.start()

            # If we have an observation with a specific wafer, modify the detector
            # selection dictionary to include only that wafer.
            obs_dets_select = dets_select
            if obs_props[obindx]["wafer"] != "all":
                if obs_dets_select is None:
                    obs_dets_select = dict()
                obs_dets_select[self.ax_detinfo_wafer_key] = obs_props[obindx]["wafer"]

            # One process in the group loads the metadata, builds the focalplane
            # model, and broadcasts to the rest of the group.
            #
            # FIXME: in the case of multiple wafers per observation, each wafer
            # reader needs to do this step, in order to pull in unique metadata
            # that is per-wafer and merge.
            #
            obs_meta, det_props = self._load_metadata(
                obs_name,
                obs_props[obindx]["session_name"],
                comm.comm_group,
                obs_dets_select,
                preproc_conf,
            )

            # Create the instrument model for the observation
            telescope, fp_flags = self._create_obs_instrument(
                obs_name,
                comm.comm_group,
                det_props,
            )

            # Create the observation and allocate data objects
            ob, have_pointing = self._create_observation(
                obs_name,
                obs_props[obindx]["session_name"],
                comm,
                telescope,
                n_samp,
                fp_flags,
                obs_meta,
            )

            # Read and communicate data
            self._load_data(ob, have_pointing, preproc_conf)

            # Compute the boresight pointing and observatory position
            if have_pointing:
                compute_boresight_pointing(
                    ob,
                    self.times,
                    self.azimuth,
                    self.elevation,
                    self.roll,
                    self.boresight_azel,
                    self.boresight_radec,
                    defaults.position,
                    defaults.velocity,
                    self.shared_flags,
                    defaults.shared_mask_processing,
                )
            data.obs.append(ob)
            log.info_rank(
                f"LoadContext {obs_name} loaded in",
                comm=comm.comm_group,
                timer=otimer,
            )

    @function_timer
    def _load_metadata(self, obs_name, session_name, gcomm, dets_select, preproc_conf):
        """Load observation metadata and the focalplane properties.

        One process in the group loads the metadata, builds the focalplane
        model, and broadcasts to the rest of the group.

        Args:
            obs_name (str):  The observation name
            session_name (str):  The observing session name (context obs ID).
            gcomm (MPI.Comm):  The group communicator or None.
            dets_select (dict):  The detector selection dictionary passed to
                Context.get_meta()
            preproc_conf:  If not None, the preprocessing config dictionary used
                to cut detectors when loading.

        Returns:
            (tuple):  The (observation metadata, detector property table)
                for the observation.

        """
        log = Logger.get()
        timer = Timer()
        timer.start()
        if gcomm is None:
            rank = 0
        else:
            rank = gcomm.rank

        det_props = None
        obs_meta = None

        # FIXME: This only works if there is one wafer per observation (and
        # hence one reader).
        if rank == 0:
            # Load metadata
            ctx = open_context(context=self.context, context_file=self.context_file)
            meta = ctx.get_meta(session_name, dets=dets_select)
            if self.context_file is not None:
                del ctx

            # Parse the axis manager metadata into observation metadata
            # and detector properties.
            obs_meta = dict()
            fp_cols = dict()
            parse_metadata(
                meta,
                obs_meta,
                fp_cols,
                self.ax_pathsep,
                self.axis_detector,
                None,
                None,
            )

            # Set the column used by toast as the detector name.
            fp_cols["name"] = Column(
                name="name",
                data=fp_cols[f"det_info{self.ax_pathsep}readout_id"].data,
            )

            if self.analytic_bandpass:
                # Add bandpass information to the focalplane
                try:
                    band = fp_cols[f"det_info{self.ax_pathsep}band"].data
                except KeyError:
                    band = fp_cols[
                        f"det_info{self.ax_pathsep}wafer{self.ax_pathsep}bandpass"
                    ].data
                freq = [float(b[1:]) for b in band]
                bandcenter = np.array(freq) * u.GHz
                bandwidth = bandcenter * self.bandwidth
                fp_cols["bandcenter"] = Column(name="bandcenter", data=bandcenter)
                fp_cols["bandwidth"] = Column(name="bandwidth", data=bandwidth)

            # Construct table
            det_props = QTable(fp_cols)
            del meta

            # Detector ordering.  When distributing and reading data, we
            # need to have the detectors sorted into contiguous blocks of
            # of detectors.  We sort the table now by wafer and then by name.
            wafer_key = f"det_info{self.ax_pathsep}{self.ax_detinfo_wafer_key}"
            det_props.sort([wafer_key, "name"])

        log.debug_rank(
            f"LoadContext {obs_name} metadata loaded in",
            comm=gcomm,
            timer=timer,
        )

        if gcomm is not None:
            obs_meta = gcomm.bcast(obs_meta, root=0)
            det_props = gcomm.bcast(det_props, root=0)

        log.debug_rank(
            f"LoadContext {obs_name} metadata bcast took",
            comm=gcomm,
            timer=timer,
        )
        return (obs_meta, det_props)

    @function_timer
    def _create_obs_instrument(self, obs_name, gcomm, det_props):
        """Create telescope, session, and focalplane flags.

        These are objects needed prior to instantiating the Observation.

        Args:
            obs_name (str):  The observation name
            gcomm (MPI.Comm):  The group communicator or None.
            det_props (QTable):  The astropy table of detector properties.

        Returns:
            (tuple):  The (Telescope, per-detector flags) where the flags
                cut detectors with the processing bit if they have no
                pointing reconstruction.

        """
        log = Logger.get()

        # Convert any focalplane quaternion offsets to toast format.  We also
        # look for any detectors with NaN values and flag those with the
        # processing bit.

        xi_key = f"focal_plane{self.ax_pathsep}xi"
        eta_key = f"focal_plane{self.ax_pathsep}eta"
        gamma_key = f"focal_plane{self.ax_pathsep}gamma"
        wafer_type_key = f"det_info{self.ax_pathsep}wafer{self.ax_pathsep}type"

        fp_flags = dict()
        if xi_key in det_props.colnames:
            fp_bad = np.logical_or(
                np.isnan(det_props[xi_key]),
                np.logical_or(
                    np.isnan(det_props[eta_key]),
                    np.isnan(det_props[gamma_key]),
                ),
            )
            fp_flags = {
                x: defaults.det_mask_processing
                for x, y in zip(det_props["name"], fp_bad)
                if y
            }
            det_props[xi_key][fp_bad] = 0
            det_props[eta_key][fp_bad] = 0
            det_props[gamma_key][fp_bad] = 0
            quat_data = toast.instrument_coords.xieta_to_quat(
                det_props[xi_key],
                det_props[eta_key],
                det_props[gamma_key],
            )
        else:
            # No detector offsets yet
            quat_data = np.tile(
                np.array([0, 0, 0, 1], dtype=np.float64),
                len(det_props[f"det_info{self.ax_pathsep}readout_id"]),
            ).reshape((-1, 4))

        # Flag non-optical detectors with the processing mask
        if wafer_type_key in det_props.colnames:
            fp_dark = det_props[wafer_type_key] != "OPTC"
            for dname, dark in zip(det_props["name"], fp_dark):
                if dark:
                    fp_flags[dname] |= defaults.det_mask_processing

        # Do we have any good detectors left?
        n_good = 0
        for det in det_props["name"]:
            if det not in fp_flags or fp_flags[det] == 0:
                n_good += 1
        if n_good == 0:
            log.warning_rank(
                f"LoadContext {obs_name} has no unflagged detectors, skipping!",
                comm=gcomm,
            )
            return

        # Add a column for the quaternion offset of each detector, as well
        # as the gamma angle with standard naming.
        quat_col = Column(
            name="quat",
            data=quat_data,
        )
        det_props.add_column(quat_col, index=0)
        gamma_col = Column(
            name="gamma",
            data=det_props[gamma_key] * u.radian,
        )
        det_props.add_column(gamma_col)

        focalplane = toast.instrument.Focalplane(
            detector_data=det_props,
            sample_rate=1.0 * u.Hz,
        )

        # Use the default observatory location
        raw_site = so3g.proj.coords.SITES["_default"]
        tele_lon_deg = raw_site.lon
        tele_lat_deg = raw_site.lat
        tele_alt_m = raw_site.elev

        # Construct a toast weather object.  If we had metadata containing the
        # current conditions (temperature, PWV, wind, etc) we could use it here.
        # For now, we use the EarthlySite.typical_weather so that our pointing
        # corrections are the same as other parts of the code.  For the weather
        # parameters not in this dictionary, we use the toast-bundled MERRA-2
        # data for the Atacama.  We add this MERRA-2 data later once we have the
        # observing session start time.

        site_temp_celsius = raw_site.typical_weather["temperature"]
        site_humidity = raw_site.typical_weather["humidity"]
        site_pressure = raw_site.typical_weather["pressure"]

        site_weather = toast.weather.Weather(
            time=None,
            ice_water=None,
            liquid_water=None,
            pwv=None,
            humidity=None,
            surface_pressure=site_pressure * u.mbar,
            surface_temperature=site_temp_celsius * u.Celsius,
            air_temperature=site_temp_celsius * u.Celsius,
            west_wind=None,
            south_wind=None,
        )
        site_weather.relative_humidity = site_humidity

        site = SOSite(
            name="SO",
            lat=tele_lat_deg * u.degree,
            lon=tele_lon_deg * u.degree,
            alt=tele_alt_m * u.meter,
            weather=site_weather,
        )

        telescope = toast.instrument.Telescope(
            self.telescope_name, focalplane=focalplane, site=site
        )
        return telescope, fp_flags

    @function_timer
    def _create_observation(
        self, obs_name, session_name, comm, telescope, n_samp, fp_flags, meta
    ):
        """Create the observation.

        Note that the focalplane table has already been sorted by wafer
        name and then detector name.

        Args:
            obs_name (str):  The observation name
            session_name (str):  The observing session name (context obs ID).
            gcomm (MPI.Comm):  The group communicator or None.
            telescope (Telescope):  The Telescope for this observation.
            n_samp (int):  The number of samples in the observation.
            fp_flags (dict):  The per-detector flags to apply.
            meta (dict):  The observation metadata to populate.

        Returns:
            (tuple):  The (Observation, have_pointing), where the second
                element is True if the observation has boresight pointing.

        """
        log = Logger.get()
        timer = Timer()
        timer.start()

        # Computer the detector sets
        if self.detset_key is None:
            detsets = None
        else:
            detsets = telescope.focalplane.detector_groups(self.detset_key)

        # Note:  the session times will be updated later when reading timestamps
        session = toast.instrument.Session(session_name)

        # Create the observation
        ob = toast.Observation(
            comm,
            telescope,
            n_samp,
            name=obs_name,
            session=session,
            detector_sets=detsets,
            sample_sets=None,
            process_rows=comm.group_size,
        )

        # Apply detector flags for bad pointing reconstruction
        local_dets = set(ob.local_detectors)
        local_fp_flags = {x: y for x, y in fp_flags.items() if x in local_dets}
        ob.update_local_detector_flags(local_fp_flags)

        # Create observation fields
        ob.shared.create_column(
            self.times,
            shape=(ob.n_local_samples,),
            dtype=np.float64,
        )

        fp_array = telescope.focalplane.detector_data
        ax_boresight_az = ax_name_fp_subst(self.ax_boresight_az, fp_array)
        ax_boresight_el = ax_name_fp_subst(self.ax_boresight_el, fp_array)
        ax_boresight_roll = ax_name_fp_subst(self.ax_boresight_roll, fp_array)
        ax_hwp_angle = ax_name_fp_subst(self.ax_hwp_angle, fp_array)
        ax_det_signal = ax_name_fp_subst(self.ax_det_signal, fp_array)

        have_pointing = True
        if ax_boresight_az is None:
            have_pointing = False
        else:
            ob.shared.create_column(
                self.azimuth,
                shape=(ob.n_local_samples,),
                dtype=np.float64,
            )
        if ax_boresight_el is None:
            have_pointing = False
        else:
            ob.shared.create_column(
                self.elevation,
                shape=(ob.n_local_samples,),
                dtype=np.float64,
            )
        if ax_boresight_roll is None:
            have_pointing = False
        else:
            ob.shared.create_column(
                self.roll,
                shape=(ob.n_local_samples,),
                dtype=np.float64,
            )

        if have_pointing:
            ob.shared.create_column(
                self.boresight_azel,
                shape=(ob.n_local_samples, 4),
                dtype=np.float64,
            )
            ob.shared.create_column(
                self.boresight_radec,
                shape=(ob.n_local_samples, 4),
                dtype=np.float64,
            )
        if self.hwp_angle is not None and ax_hwp_angle is not None:
            ob.shared.create_column(
                self.hwp_angle,
                shape=(ob.n_local_samples,),
                dtype=np.float64,
            )
        if self.boresight_angle is not None:
            ob.shared.create_column(
                self.boresight_angle,
                shape=(ob.n_local_samples,),
                dtype=np.float64,
            )
        if self.corotator_angle is not None:
            ob.shared.create_column(
                self.corotator_angle,
                shape=(ob.n_local_samples,),
                dtype=np.float64,
            )
        ob.shared.create_column(
            self.shared_flags,
            shape=(ob.n_local_samples,),
            dtype=np.uint8,
        )
        if have_pointing:
            ob.shared.create_column(
                defaults.position,
                shape=(ob.n_local_samples, 3),
                dtype=np.float64,
            )
            ob.shared.create_column(
                defaults.velocity,
                shape=(ob.n_local_samples, 3),
                dtype=np.float64,
            )
        if ax_det_signal is not None:
            ob.detdata.create(
                self.det_data, dtype=np.float64, units=self.det_data_units
            )
            ob.detdata.create(self.det_flags, dtype=np.uint8)

        if meta is not None:
            ob.update(meta)

        log.debug_rank(
            f"LoadContext {obs_name} allocate Observation in",
            comm=comm.comm_group,
            timer=timer,
        )
        return ob, have_pointing

    @function_timer
    def _load_data(self, ob, have_pointing, pconf):
        """Recursively load AxisManager data.

        The first process with detectors from a given wafer will read
        all data for that wafer.  Some examples:

        - 1 processes, 7 wafers:  rank 0 reads all wafers
        - 2 processes, 7 wafers:  rank 0 reads 3 wafers, rank 1 reads 4 wafers
        - 16 processes, 7 wafers:  ranks 0, 2, 4, 6, 9, 11, 13 read one wafer each

        Each reading process will optionally apply the preprocessing
        to the loaded data prior to distribution. The rank zero process (which
        is always a reader) extracts any shared data and additional metadata that
        is found.  Reading processes communicate slices of detector data to all
        other processes.

        Args:
            ob (Observation):  The observation to populate.
            have_pointing (bool):  True if the data contains boresight pointing.
            pconf (dict):  The optional preprocessing config.

        Returns:
            None

        """
        log = Logger.get()
        timer = Timer()
        timer.start()

        comm = ob.comm
        gcomm = comm.comm_group
        rank = ob.comm.group_rank

        # Compute the data distribution of detectors
        wafer_key = f"det_info{self.ax_pathsep}{self.ax_detinfo_wafer_key}"
        wafer_dets, wafer_readers, wafer_proc_dets, proc_wafer_dets = (
            distribute_detector_props(ob, wafer_key)
        )

        # Load data and apply preprocessing.
        axwafers = read_and_preprocess_wafers(
            ob.name,
            ob.session.name,
            gcomm,
            wafer_readers,
            wafer_dets,
            preconfig=pconf,
            ignore_preprocess_archive=self.ignore_preprocess_archive,
            context=self.context,
            context_file=self.context_file,
        )

        log.debug_rank(
            f"LoadContext {ob.name} AxisManager data loaded in",
            comm=gcomm,
            timer=timer,
        )

        # Track the fields we are extracting
        fp_array = ob.telescope.focalplane.detector_data
        ax_times = ax_name_fp_subst(self.ax_times, fp_array)
        ax_boresight_az = ax_name_fp_subst(self.ax_boresight_az, fp_array)
        ax_boresight_el = ax_name_fp_subst(self.ax_boresight_el, fp_array)
        ax_boresight_roll = ax_name_fp_subst(self.ax_boresight_roll, fp_array)
        ax_hwp_angle = ax_name_fp_subst(self.ax_hwp_angle, fp_array)
        ax_det_signal = ax_name_fp_subst(self.ax_det_signal, fp_array)
        ax_flags = list()
        for axname, bit in self.ax_flags:
            full_name = ax_name_fp_subst(axname, fp_array)
            ax_flags.append((full_name, bit))
        ax_det_flags = list()
        for axname, bit in self.ax_det_flags:
            full_name = ax_name_fp_subst(axname, fp_array)
            ax_det_flags.append((full_name, bit))

        shared_ax_to_obs = {ax_times: self.times}
        if have_pointing:
            shared_ax_to_obs[ax_boresight_az] = self.azimuth
            shared_ax_to_obs[ax_boresight_el] = self.elevation
            shared_ax_to_obs[ax_boresight_roll] = self.roll
        if self.hwp_angle is not None and ax_hwp_angle is not None:
            shared_ax_to_obs[ax_hwp_angle] = self.hwp_angle
        shared_flag_invert = {x[0]: (x[1] < 0) for x in ax_flags}
        shared_flag_fields = {x[0]: abs(x[1]) for x in ax_flags}
        det_flag_invert = {x[0]: (x[1] < 0) for x in ax_det_flags}
        det_flag_fields = {x[0]: abs(x[1]) for x in ax_det_flags}

        # The results of the recursive parsing
        extra_meta = dict()
        shared_data = dict()
        det_data = dict()
        interval_data = dict()

        # Recursively parse axis managers and build list of data to
        # populate in the observation.  Since all wafers have the same
        # ancil data (and the same field names, just with different
        # detectors), we can just do this on one process and broadcast
        # the result.
        #
        # FIXME:  In the case of multiple wafers (and readers) per
        # observation, every reader must parse every wafer and merge
        # all the metadata information.
        #
        temp_shared = None
        restricted_samps = None
        ax_shift = None
        if ob.comm.group_rank == 0:
            first_wafer = list(axwafers.keys())[0]
            self._parse_data(
                ob,
                axwafers[first_wafer],
                shared_ax_to_obs,
                shared_flag_fields,
                shared_flag_invert,
                det_flag_fields,
                det_flag_invert,
                extra_meta,
                shared_data,
                det_data,
                interval_data,
                None,
            )
            temp_shared = {x: None for x, y in shared_data.items()}

            # Does the axis manager have a truncated number of samples?
            restricted_samps = axwafers[first_wafer][self.axis_sample].count
            if restricted_samps != ob.n_local_samples:
                ax_shift = axwafers[first_wafer][self.axis_sample].offset
            else:
                ax_shift = 0

        if gcomm is not None:
            extra_meta = gcomm.bcast(extra_meta, root=0)
            temp_shared = gcomm.bcast(temp_shared, root=0)
            if ob.comm.group_rank != 0:
                shared_data = temp_shared
            det_data = gcomm.bcast(det_data, root=0)
            interval_data = gcomm.bcast(interval_data, root=0)
            restricted_samps = gcomm.bcast(restricted_samps, root=0)
            ax_shift = gcomm.bcast(ax_shift, root=0)

        # Add extra metadata that was discovered.
        ob.update(extra_meta)

        log.debug_rank(
            f"LoadContext {ob.name} AxisManager field parsing took",
            comm=gcomm,
            timer=timer,
        )

        # Create the intervals
        for intr_name, sampspans in interval_data.items():
            ob.intervals[intr_name] = toast.intervals.IntervalList(
                ob.shared[self.times], samplespans=sampspans
            )

        # Collectively store shared data.  All readers have a full copy of
        # this data, but we only set this from rank zero.  If there are
        # cut samples at the beginning and end, ensure that timestamps are
        # always valid.
        for shr_obs_name, shrbuf in shared_data.items():
            bf = shrbuf
            if shr_obs_name == self.times and restricted_samps != ob.n_local_samples:
                if rank == 0:
                    msg = f"Axman samples {restricted_samps} != {ob.n_local_samples}"
                    msg += ", extrapolating timestamps"
                    log.debug(msg)
                    bf = np.zeros(ob.n_local_samples, dtype=np.float64)
                    (rate, dt, _, _, _) = toast.utils.rate_from_times(shrbuf)
                    bf[ax_shift : ax_shift + restricted_samps] = shrbuf
                    bf[0:ax_shift] = shrbuf[0] + dt * np.arange(
                        -ax_shift, 0, 1, dtype=np.float64
                    )
                    end_gap = ob.n_local_samples - restricted_samps - ax_shift
                    bf[ax_shift + restricted_samps :] = shrbuf[-1] + dt * np.arange(
                        1, end_gap + 1, 1, dtype=np.float64
                    )
            ob.shared[shr_obs_name].set(bf, fromrank=0)

        log.debug_rank(
            f"LoadContext {ob.name} Shared data copy took",
            comm=gcomm,
            timer=timer,
        )

        # Now that we have timestamps loaded, update our focalplane sample rate
        # and observing session times.
        (rate, dt, dt_min, dt_max, dt_std) = toast.utils.rate_from_times(
            ob.shared[self.times].data
        )
        ob.telescope.focalplane.sample_rate = rate * u.Hz

        ob.session.start = datetime.fromtimestamp(
            ob.shared[self.times].data[0]
        ).astimezone(timezone.utc)
        ob.session.end = datetime.fromtimestamp(
            ob.shared[self.times].data[-1]
        ).astimezone(timezone.utc)

        # Update site weather, now that we have the observing time.
        sim_atacama = toast.weather.SimWeather(
            time=ob.session.start,
            name="atacama",
            median_weather=True,
        )
        old_weather = ob.telescope.site.weather
        new_weather = toast.weather.Weather(
            time=sim_atacama.time,
            ice_water=sim_atacama.ice_water,
            liquid_water=sim_atacama.liquid_water,
            pwv=sim_atacama.pwv,
            humidity=sim_atacama.humidity,
            surface_pressure=old_weather.surface_pressure,
            surface_temperature=old_weather.surface_temperature,
            air_temperature=old_weather.air_temperature,
            west_wind=sim_atacama.west_wind,
            south_wind=sim_atacama.south_wind,
        )
        new_weather.relative_humidity = old_weather.relative_humidity
        ob.telescope.site.weather = new_weather

        log.debug_rank(
            f"LoadContext {ob.name} sample rate calculation took",
            comm=gcomm,
            timer=timer,
        )

        # Distribute detector data
        for field, merge_list in det_data.items():
            for ax_field, ax_dtype, mask in merge_list:
                do_invert = False
                if ax_field in det_flag_invert:
                    do_invert = det_flag_invert[ax_field]
                distribute_detector_data(
                    ob,
                    field,
                    axwafers,
                    self.axis_detector,
                    self.axis_sample,
                    ax_field,
                    ax_dtype,
                    wafer_readers,
                    wafer_proc_dets,
                    proc_wafer_dets,
                    str(self.ax_pathsep),
                    is_flag=(mask is not None),
                    flag_invert=do_invert,
                    flag_mask=mask,
                )

        # Original wafer data no longer needed.  AxisManager does not seem to
        # have a clean destructor, so do it manually.
        def _ax_del_children(ax):
            for k in list(ax.keys()):
                if isinstance(ax[k], AxisManager):
                    _ax_del_children(ax[k])
                del ax[k]

        for wfname in list(axwafers.keys()):
            _ax_del_children(axwafers[wfname])
            del axwafers[wfname]
        del axwafers

        log.debug_rank(
            f"LoadContext {ob.name} Detector data distribution took",
            comm=gcomm,
            timer=timer,
        )

    def _parse_data(
        self,
        obs,
        axman,
        shared_ax_to_obs,
        shared_flag_fields,
        shared_flag_invert,
        det_flag_fields,
        det_flag_invert,
        extra_meta,
        shared_data,
        det_data,
        interval_data,
        base,
    ):
        """Recursively parse AxisManager data products.

        This (mostly) does not actually extract the data, but instead builds
        up information needed to extract that data in following steps.

        Args:
            obs (Observation):  The observation
            axman (AxisManager):  The top-level axis manager
            shared_ax_to_obs (dict):  For each ancil axis manager field, the
                corresponding observation shared key.
            shared_flag_fields (dict):  For each shared axis manager field, the
                (target observation flag field, bit mask) where this should be
                merged.
            shared_flag_invert (dict):  For each shared axis manager flag field,
                whether the meaning of the flag should be inverted.
            det_flag_fields (dict):  For each axis manager detector flag field,
                the (target observation detdata field, bit mask) where this
                should be merged.
            det_flag_invert (dict):  For each axis manager detector flag field,
                whether the meaning of the flag should be inverted.
            extra_meta (dict):  Any new / additional metadata found on the root
                process.
            shared_data (dict):  For each shared observation field, a handle to
                the axis manager data buffer to use.  This is only significant
                on the rank zero process.
            det_data (dict):  For each detector data field, a tuple of (axman
                field, flag mask).  This is significant on the reading processes.
                If the data is signal then flag mask will be None.
            interval_data (dict):  For each observation interval field, the
                sample spans to use.  Only significant on the rank zero process.
            base (str):  The current dictionary key for this recursion level.

        Returns:
            None

        """

        def _valid_meta(obj):
            # Check that the object is an allowed type before adding it to the metadata
            if np.isscalar(obj):
                return True
            if isinstance(obj, np.ndarray):
                return True
            if isinstance(obj, (list, set, dict, tuple, str)):
                return True
            return False

        # Find the per-observation names we might be using
        fp_array = obs.telescope.focalplane.detector_data
        ax_det_signal = ax_name_fp_subst(self.ax_det_signal, fp_array)

        # Some metadata has already been parsed, but some new values
        # may only show up when reading data, so we need to handle those
        # as well.
        rank = obs.comm.group_rank

        created_meta_extra = False
        created_meta_obs = False

        if base is None:
            mcur = obs
            mext = extra_meta
        else:
            if base not in obs:
                created_meta_obs = True
                obs[base] = dict()
            mcur = obs[base]
            if base not in extra_meta:
                created_meta_extra = True
                extra_meta[base] = dict()
            mext = extra_meta[base]
        for key in axman.keys():
            if base is not None:
                data_key = f"{base}{self.ax_pathsep}{key}"
            else:
                data_key = key
            if isinstance(axman[key], AxisInterface):
                # This is one of the axes
                continue
            if isinstance(axman[key], FlagManager):
                # Descend- Anything different we need to do here?
                self._parse_data(
                    obs,
                    axman[key],
                    shared_ax_to_obs,
                    shared_flag_fields,
                    shared_flag_invert,
                    det_flag_fields,
                    det_flag_invert,
                    extra_meta,
                    shared_data,
                    det_data,
                    interval_data,
                    key,
                )
            elif isinstance(axman[key], AxisManager):
                # Descend
                self._parse_data(
                    obs,
                    axman[key],
                    shared_ax_to_obs,
                    shared_flag_fields,
                    shared_flag_invert,
                    det_flag_fields,
                    det_flag_invert,
                    extra_meta,
                    shared_data,
                    det_data,
                    interval_data,
                    key,
                )
            else:
                # FIXME:  It would be nicer if this information was available
                # through a public member...
                field_axes = axman._assignments[key]
                if len(field_axes) == 0:
                    # This data is not associated with an axis.  If it does not
                    # yet exist in the observation metadata, then add it.
                    if _valid_meta(axman[key]):
                        mext[key] = axman[key]
                elif field_axes[0] == self.axis_detector:
                    if len(field_axes) == 1:
                        # This is a detector property
                        if data_key in obs.telescope.focalplane.detector_data.colnames:
                            # We already included this in the detector properties
                            continue
                        # This must be some per-detector derived data- add to the
                        # observation dictionary
                        if data_key not in mcur:
                            if _valid_meta(axman[key]):
                                mext[key] = axman[key]
                    elif field_axes[1] == self.axis_sample:
                        # This is detector data.  See if it is one of the standard
                        # fields we are parsing.
                        if hasattr(axman[key], "dtype"):
                            # This is an array
                            dt = axman[key].dtype
                        else:
                            # This is a RangesMatrix of flags
                            dt = np.dtype(np.uint8)
                        if data_key == ax_det_signal:
                            # Detector signal
                            det_data[self.det_data] = [(data_key, dt, None)]
                        elif data_key in det_flag_fields:
                            # One of the flag fields
                            if self.det_flags not in det_data:
                                det_data[self.det_flags] = list()
                            det_data[self.det_flags].append(
                                (
                                    data_key,
                                    dt,
                                    det_flag_fields[data_key],
                                )
                            )
                        else:
                            # Some other kind of detector data.  Ignore this for now,
                            # since this needs to be pre-created in the observation
                            # for all processes (not just the readers).  If we need
                            # to read other detector data fields we should add a trait
                            # for those specifying the shape and type.
                            pass
                    else:
                        # Must be some other type of object...
                        if data_key not in mcur:
                            if _valid_meta(axman[key]):
                                mext[key] = axman[key]
                elif field_axes[0] == self.axis_sample:
                    # This is shared data
                    if isinstance(axman[key], so3g.proj.Ranges):
                        # This is a set of 1D shared ranges.  Translate this to a
                        # toast interval list.
                        if rank == 0:
                            samplespans = list()
                            for rg in axman[key].ranges():
                                samplespans.append((rg[0], rg[1] - 1))
                            interval_data[data_key] = samplespans
                    elif data_key in shared_ax_to_obs:
                        # One of our selected shared data fields
                        if rank == 0:
                            shared_data[shared_ax_to_obs[data_key]] = axman[key]
                    elif data_key in shared_flag_fields:
                        # One of our selected flag fields
                        if obs.rank == 0:
                            if self.shared_flags not in shared_data:
                                shared_data[self.shared_flags] = np.array(
                                    obs.shared[self.shared_flags].data
                                )
                            temp = shared_flag_fields[data_key] * np.ones_like(
                                axman[key]
                            )
                            if shared_flag_invert[data_key]:
                                temp[axman[key] != 0] = 0
                            else:
                                temp[axman[key] == 0] = 0
                            shared_data[self.shared_flags] |= temp
                            del temp
                    else:
                        # This is some other shared data.
                        # FIXME: we skip this for now.
                        continue
                else:
                    # Some other object...
                    if data_key not in mcur:
                        if _valid_meta(axman[key]):
                            mext[key] = axman[key]

        # Clean up any dictionaries that we created if they were empty
        if created_meta_extra and len(mext) == 0:
            del extra_meta[base]
        if created_meta_obs and len(mcur) == 0:
            del obs[base]

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        return dict()

    def _provides(self):
        prov = {
            "shared": [
                self.times,
                self.boresight_azel,
            ],
            "detdata": [self.det_data],
        }
        if self.boresight_radec is not None:
            prov["shared"].append(self.boresight_radec)
        if self.boresight_angle is not None:
            prov["shared"].append(self.boresight_angle)
        if self.corotator_angle is not None:
            prov["shared"].append(self.corotator_angle)
        if self.hwp_angle is not None:
            prov["shared"].append(self.hwp_angle)
        return prov



import os
from urllib.request import urlretrieve

from ast import literal_eval

# This is a listing of public URLs from which to grab some useful data
# files. On shared systems it may be wise to maintain a shared cache of
# such files -- see get_local_file.

RESOURCE_DEFAULTS = {
    "de421.bsp": "ftp://ssd.jpl.nasa.gov/pub/eph/planets/bsp/de421.bsp",
}


def get_local_file(filename: str, cache: bool = True, download: bool = True) -> str:
    """This function utilizes RESOURCE_DEFAULTS or SOTODLIB_RESOURCES environment
    variable to manage resource files such as a planet catalog from NASA.

    RESOURCE_DEFAULTS and SOTODLIB_RESOURCES are dictionaries with key a filename
    and value either an FTP or absolute path to the file. The FTP path starts with
    "ftp://" and an absolute path with "file://". For example, a valid value of
    RESOURCE_DEFAULTS or SOTODLIB_RESOURCES is:

    {
    "de421.bsp": "ftp://ssd.jpl.nasa.gov/pub/eph/planets/bsp/de421.bsp",
    "de422.bsp": "file:///scratch/gpfs/SIMONSOBS/de422.bsp"
    }

    **Note**: When setting SOTODLIB_RESOURCES use a serialized version of the
              dictionary.

    Upon call, this function checks for the filename first in SOTODLIB_RESOURCES,
    and then at the RESOURCE_DEFAULTS. When the key value is an FTP URL, the
    function checks for the file in a predefined path under the user's home
    folder, and returns its path when it exists, otherwise it downloads it.
    If the key value is a path it just returns it, and it assumes that the user
    ahs verified that the file exists.

    The parameter cache sets the folder of the where a file will be downloaded.
    When True, files are stored under the user's home folder. When cache is set
    to False, the files are downloaded under the tmp filesystem and their long
    term storage is not guaranteed.

    Args:
      filename: The name of the file to grab
      cache: A boolean indicating that downloaded files should be cached in user
              home folder (~/.sotodlib/filecache/).
      download: If False, the file will not be downloaded even if
        config specifies a URL and the local cache does not have the file.

    Returns:
      The absolute path of the file.

    Exceptions:
        RuntimeError: when the requested resource file does not exist as a key
                      in SOTODLIB_RESOURCES env variable or RESOURCE_DEFAULTS.
        RuntimeErorr: When the value of a key is not an ftp or file path.

    """

    # Local cache. This is per user, however we may want to try and right in
    # the path this file is first and then the user's home directory.
    local_cache = os.path.join(os.path.expanduser("~"), ".sotodlib/filecache/")

    env_resource_paths = literal_eval(os.environ.get("SOTODLIB_RESOURCES", "{}"))

    de_url = env_resource_paths.get(filename, None)

    if de_url is None:
        de_url = RESOURCE_DEFAULTS.get(filename, None)

    if de_url is None:
        raise RuntimeError(
            f"File {filename} not registered in RESOURCE_DEFAULTS nor in "
            + "$SOTODLIB_RESOURCES. Please ensure the URL to the file is "
            + "in RESOURCE_DEFAULTS, and ask your system maintainer "
            + "to expose a cached copy through $SOTODLIB_RESOURCES."
        )

    if de_url.startswith("ftp://"):
        # Check if file is already there.
        file_cached = os.path.exists(os.path.join(local_cache, filename))

        if file_cached:
            return os.path.join(local_cache, filename)

        # Set the target path for the file and create it if it does not exist
        target_path = local_cache if cache else "/tmp/"
        os.makedirs(name=target_path, exist_ok=True)
        target_file = os.path.join(target_path, filename)
        if download:
            _, headers = urlretrieve(de_url, target_file)
    elif de_url.startswith("file://"):
        target_file = de_url[7:]
    else:
        raise RuntimeError(
            f"Malformed resource file URL {de_url}. Resource file URL needs to "
            + 'start with "ftp://" or "file://"'
        )

    return target_file



"""
Script for running updates on (or creating) a g3tsmurf database. This setup
is specifically designed to work when the data is dynamically coming in. Meaning is 
is designed to work from something like a cronjob. 
"""
import os
import yaml
import datetime as dt
import numpy as np
import argparse
import logging
from sqlalchemy import not_, or_, and_
from typing import Optional

from sotodlib.io.load_smurf import G3tSmurf, Observations, logger
from sotodlib.io.datapkg_utils import load_configs


def main(config: Optional[str] = None, update_delay: float = 2, 
         from_scratch: bool = False, verbosity: int = 2,
         index_via_actions: bool=False, checked_file: Optional[str]=None):
    """
    Arguments
    ---------
    config: string
        configuration file for G3tSmurf
    update_delay: float
        number of days to 'look back' to update observation information
    from_scratch: bool
        if True, run database update with minimum ctime of 1.6e9 (all SO time).
        overrides update_delay
    verbosity: int
        0-3, higher numbers = more printouts
    index_via_actions: bool
        if True, will look through action folders to create observations, this
        will be necessary for data older than Oct 2022 but creates concurancy 
        issues on systems (like the site) running automatic deletion of level 2 
        data.
    checked_file: str
        a file name that contains a list of observations that would by default 
        cause errors to be thrown during this script but have been manually
        checked and dealt with
    """
    show_pb = True if verbosity > 1 else False

    if verbosity == 0:
        logger.setLevel(logging.ERROR)
    elif verbosity == 1:
        logger.setLevel(logging.WARNING)
    elif verbosity == 2:
        logger.setLevel(logging.INFO)
    elif verbosity == 3:
        logger.setLevel(logging.DEBUG)

    if from_scratch:
        logger.info("Building Database from Scratch, May take awhile")
        min_time = dt.datetime.utcfromtimestamp(int(1.6e9))
        make_db = True
    else:
        min_time = dt.datetime.now() - dt.timedelta(days=update_delay)
        make_db = False

    cfgs = load_configs( config )
    SMURF = G3tSmurf.from_configs(cfgs, make_db=make_db)

    updates_start = dt.datetime.now().timestamp()

    session = SMURF.Session()
    SMURF.index_metadata(min_ctime=min_time.timestamp(), session=session)
    SMURF.index_archive(
        min_ctime=min_time.timestamp(), 
        show_pb=show_pb, 
        session=session
    )
    if index_via_actions:
        SMURF.index_action_observations(
            min_ctime=min_time.timestamp(),
            session=session
        )    
    SMURF.index_timecodes(
        min_ctime=min_time.timestamp(),
        session=session
    )
    SMURF.update_finalization(update_time=updates_start, session=session)
    SMURF.last_update = updates_start

    new_obs = session.query(Observations).filter(
        or_(
            Observations.start >= min_time,
            Observations.start == None,
        )
    ).all()

    raise_list_timing = []
    raise_list_readout_ids = []
    for obs in new_obs:
        if obs.stop is None or len(obs.tunesets)==0:
            SMURF.update_observation_files(
                obs, 
                session, 
                force=True,
            )
        if (obs.stop is not None) and (not obs.timing):
            raise_list_timing.append(obs.obs_id)

        if (obs.stop is not None) and len(obs.tunesets)==0:
            raise_list_readout_ids.append(obs.obs_id)

    if len(raise_list_timing) > 0 or len(raise_list_readout_ids) > 0:
        logger.info(
            f"Found observations with bad timing or missing readout ids "
            "checking to see if they've been manually cleared"
        )
        if checked_file is None or not os.path.exists(checked_file):
            logger.warning(
                f"File {checked_file} does not exist so cannot check if "
                "problematic observations have been manually cleared"
            )
            cleared = []
        else:
            with open(checked_file) as f:
                cleared = [c.strip("\n").strip() for c in f.readlines()]
        raise_list_timing = [x for x in raise_list_timing if x not in cleared]
        raise_list_readout_ids = [
            x for x in raise_list_readout_ids if x not in cleared
        ]
    if len(raise_list_timing) > 0 or len(raise_list_readout_ids) > 0:
        raise ValueError(
            f"Found {len(raise_list_timing)} observations with bad timing"
            f" obs_ids are {raise_list_timing}.\nFound "
            f"{len(raise_list_readout_ids)} observations without Tunesets" 
            f" obs_ids are {raise_list_readout_ids}."
        )

def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument('config', help="g3tsmurf db configuration file")
    parser.add_argument('--update-delay', help="Days to subtract from now to set as minimum ctime",
                        default=2, type=float)
    parser.add_argument('--from-scratch', help="Builds or updates database from scratch",
                        action="store_true")
    parser.add_argument("--verbosity", help="increase output verbosity. 0:Error, 1:Warning, 2:Info(default), 3:Debug",
                        default=2, type=int)
    parser.add_argument('--index-via-actions', help="Look through action folders to create observations",
                        action="store_true")
    parser.add_argument("--checked-file", 
        help="Filename of file containing a list of observations that are problematic but have been manually acknowledged")
    return parser

if __name__ == '__main__':
    parser = get_parser(parser=None)
    args = parser.parse_args()
    main(**vars(args))
